import typing

from ascetic_ddd.faker.domain.specification.interfaces import ISpecificationVisitor, ISpecification

__all__ = ("EmptySpecification",)


class EmptySpecification(ISpecification):

    __slots__ = ('_hash',)

    def __init__(self):
        self._hash: int | None = None

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = hash(frozenset())
        return self._hash

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EmptySpecification)

    def is_satisfied_by(self, obj: typing.Any) -> bool:
        return True

    def accept(self, visitor: ISpecificationVisitor):
        visitor.visit_empty_specification()



import typing
from collections.abc import Awaitable, Callable

from .interfaces import IDisposable

__all__ = ("Disposable",)


class Disposable(IDisposable):
    _callback: Callable[[], Awaitable[None]]

    def __init__(self, callback: Callable[[], Awaitable[None]]) -> None:
        self._callback = callback

    async def dispose(self) -> None:
        await self._callback()

    def __add__(self, other: IDisposable):
        return CompositeDisposable((self, other))


class CompositeDisposable(IDisposable):
    def __init__(self, delegates: typing.Iterable[IDisposable]) -> None:
        self._delegates = delegates

    async def dispose(self) -> None:
        for delegate in self._delegates:
            await delegate.dispose()

    def __add__(self, other: IDisposable):
        return CompositeDisposable((*tuple(self._delegates), other))



import typing

from ascetic_ddd.seedwork.domain.identity.interfaces import IAccessible
from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.faker.domain.specification.interfaces import ISpecification
from ascetic_ddd.seedwork.domain.utils.data import hashable
from ascetic_ddd.observable.observable import Observable

__all__ = ('InMemoryRepository',)


T = typing.TypeVar("T", covariant=True)


class InMemoryRepository(Observable, typing.Generic[T]):
    _id_attr: str = None
    _aggregates: dict[T]
    _agg_exporter: typing.Callable[[T], dict]

    def __init__(self, agg_exporter: typing.Callable[[T], dict], id_attr: str = None):
        super().__init__()
        self._agg_exporter = agg_exporter
        self._id_attr = id_attr
        self._aggregates = dict()

    async def insert(self, session: ISession, agg: T):
        state = self._agg_exporter(agg)
        id_ = self._id(state)
        self._aggregates[hashable(id_)] = agg
        await self.anotify('inserted', session, agg)

    async def update(self, session: ISession, agg: T):
        state = self._agg_exporter(agg)
        id_ = self._id(state)
        self._aggregates[hashable(id_)] = agg
        await self.anotify('updated', session, agg)

    async def get(self, session: ISession, id_: IAccessible[typing.Any] | typing.Any) -> T | None:
        key = id_.value if hasattr(id_, 'value') else id_
        return self._aggregates.get(hashable(key))

    async def find(self, session: ISession, specification: ISpecification) -> typing.Iterable[T]:
        """Add index support by extending the class"""
        for id_, agg in self._aggregates.items():
            if specification.is_satisfied_by(agg):
                yield agg

    async def setup(self, session: ISession):
        pass

    async def cleanup(self, session: ISession):
        pass

    def _id(self, state: dict) -> typing.Any:
        if self._id_attr is not None:
            return state.get(self._id_attr)
        return next(iter(state.values()))

    def __copy__(self):
        return self

    def __deepcopy__(self, memodict={}):
        return self



import typing
import weakref
from contextlib import asynccontextmanager
from time import perf_counter
from types import TracebackType
# from psycopg import AsyncConnection
from psycopg import IsolationLevel

from ascetic_ddd.observable.observable import Observable
from ...domain.session.interfaces import ISessionPool, ISession
from ..session.interfaces import (
    IPgSession, IIdentityMap, IAsyncConnection, IAsyncConnectionPool, IAsyncCursor, Query, Params
)
from .identity_map import IdentityMap

__all__ = (
    "PgSession",
    "PgSessionPool",
    "PgTransactionSession",
    "PgSavepointSession",
    "extract_connection",
    "AsyncCursorStatsDecorator",
    "AsyncConnectionStatsDecorator",
)


def extract_connection(session: IPgSession) -> IAsyncConnection:
    return session.connection


class PgSessionPool(Observable, ISessionPool):
    _pool: IAsyncConnectionPool

    def __init__(self, pool: IAsyncConnectionPool) -> None:
        self._pool = pool
        super().__init__()

    @asynccontextmanager
    async def session(self) -> typing.AsyncIterator[ISession]:
        async with self._pool.connection() as conn:
            # await conn.set_isolation_level(IsolationLevel.READ_COMMITTED)
            session = self._make_session(conn)
            await self.anotify(
                aspect='session_started',
                session=session
            )
            try:
                yield session
            finally:
                await self.anotify(
                    aspect='session_ended',
                    session=session
                )

    @staticmethod
    def _make_session(connection):
        return PgSession(connection)


class PgSession(Observable, IPgSession):
    _connection: IAsyncConnection
    _parent: typing.Optional["PgSession"]
    _identity_map: IIdentityMap

    def __init__(self, connection: IAsyncConnection, parent: typing.Optional["PgSession"] = None):
        # self._connection = connection
        self._connection = AsyncConnectionStatsDecorator(connection, self)
        self._parent = parent
        self._identity_map = IdentityMap(isolation_level=IdentityMap.READ_UNCOMMITTED)
        super().__init__()

    @property
    def connection(self) -> IAsyncConnection:
        return self._connection

    @property
    def identity_map(self) -> IIdentityMap:
        return self._identity_map

    @asynccontextmanager
    async def atomic(self) -> typing.AsyncIterator[ISession]:
        async with self.connection.transaction() as transaction:
            transaction_session = self._make_transaction_session(transaction.connection)
            await self.anotify(
                aspect='session_started',
                session=transaction_session
            )
            try:
                yield transaction_session
            finally:
                await self.anotify(
                    aspect='session_ended',
                    session=transaction_session
                )

    def _make_transaction_session(self, connection):
        return PgTransactionSession(connection, IdentityMap(), self)


# TODO: Add savepoint support
class PgTransactionSession(PgSession):

    def __init__(
            self,
            connection: IAsyncConnection,
            identity_map: IIdentityMap,
            parent: typing.Optional["PgSession"] = None
    ):
        super().__init__(connection, parent)
        self._identity_map = identity_map

    @asynccontextmanager
    async def atomic(self) -> typing.AsyncIterator[ISession]:
        savepoint_session = self._make_savepoint_session(self._connection)
        await self.anotify(
            aspect='session_started',
            session=savepoint_session
        )
        try:
            yield savepoint_session
        finally:
            await self.anotify(
                aspect='session_ended',
                session=savepoint_session
            )

    @asynccontextmanager
    async def atomic2(self) -> typing.AsyncIterator[ISession]:
        async with self.connection.transaction() as transaction:
            savepoint_session = self._make_savepoint_session(transaction.connection)
            await self.anotify(
                aspect='session_started',
                session=savepoint_session
            )
            try:
                yield savepoint_session
            finally:
                await self.anotify(
                    aspect='session_ended',
                    session=savepoint_session
                )

    def _make_savepoint_session(self, connection):
        return PgSavepointSession(connection, self._identity_map, self)


class PgSavepointSession(PgTransactionSession):
    pass


class AsyncCursorStatsDecorator:
    _delegate: IAsyncCursor
    _session: weakref.ReferenceType[IPgSession]

    def __init__(self, delegate: IAsyncCursor, session: IPgSession):
        self._delegate = delegate
        self._session = weakref.ref(session)

    async def execute(
        self,
        query: Query,
        params: Params | None = None,
        *,
        prepare: bool | None = None,
        binary: bool | None = None,
    ):
        await self._session().anotify(
            aspect='query_started',
            query=query,
            params=params,
            sender=self,
            session=self._session,
        )
        time_start = perf_counter()
        await self._delegate.execute(query, params, prepare=prepare, binary=binary)
        response_time = perf_counter() - time_start
        await self._session().anotify(
            aspect='query_ended',
            query=query,
            params=params,
            sender=self,
            session=self._session,
            response_time=response_time,
        )
        return self

    async def __aenter__(self):
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._delegate.__aexit__(exc_type, exc_val, exc_tb)

    def __getattr__(self, name):
        return getattr(self._delegate, name)


class AsyncConnectionStatsDecorator:
    _delegate: IAsyncConnection
    _session: weakref.ReferenceType[IPgSession]

    def __init__(self, delegate: IAsyncConnection, session: IPgSession):
        self._delegate = delegate
        self._session = weakref.ref(session)

    def cursor(self, *a, **kw):
        return AsyncCursorStatsDecorator(self._delegate.cursor(*a, **kw), self._session())

    async def __aenter__(self):
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._delegate.__aexit__(exc_type, exc_val, exc_tb)

    def __getattr__(self, name):
        return getattr(self._delegate, name)



import os
import time
import logging
import typing
import datetime
import subprocess

from psycopg.conninfo import conninfo_to_dict
from ascetic_ddd.faker.infrastructure.dump.interfaces import IDump, IFileDump

__all__ = ("PgDump", "GzipDump", "FileDump",)


class FileDump(IFileDump):
    _base_path: str
    _ttl: datetime.timedelta
    _delegate: IDump

    def __init__(self, delegate: IDump, base_path: str, ttl: datetime.timedelta):
        self._base_path = base_path
        self._ttl = ttl
        self._delegate = delegate
        self._logger = logging.getLogger(__name__)

    @property
    def ttl(self) -> datetime.timedelta:
        return self._ttl

    async def exists(self, name: str) -> bool:
        filename = self.make_filepath(name)
        if os.path.exists(filename):
            mtime = os.path.getmtime(filename)
            delta = time.time() - mtime
            if delta < self._ttl.total_seconds():
                self._logger.info("Dump %s exists", filename)
                return True
        self._logger.info("Dump %s does not exist", filename)
        return False

    async def dump(self, name: str):
        if not os.path.exists(self._base_path):
            os.mkdir(self._base_path)
        filename = self.make_filepath(name)
        with open(filename, "wb") as file:
            self._logger.info("Make dump %s", filename)
            return await self._delegate.dump(file)

    async def load(self, name: str):
        filename = self.make_filepath(name)
        with open(filename, "rb") as file:
            self._logger.info("Load dump %s", filename)
            return await self._delegate.load(file)

    def make_filepath(self, name: str) -> str:
        return os.path.join(self._base_path, "%s.sql.gz" % name)


class GzipDump(IDump):
    _delegate: IDump

    def __init__(self, delegate: IDump):
        self._delegate = delegate

    async def dump(self, out: typing.IO[bytes]):
        proc = subprocess.Popen(
            ["gzip"], stdin=subprocess.PIPE, stdout=out, stderr=subprocess.PIPE
        )
        await self._delegate.dump(proc.stdin)

        sub_out, sub_err = proc.communicate()
        code = proc.returncode
        if code != 0:
            raise Exception("gzip failed for Postgres: %s, %s", code, sub_err)

    async def load(self, in_: typing.IO[bytes]):
        proc = subprocess.Popen(
            ["gunzip"], stdin=in_, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        await self._delegate.load(proc.stdout)

        sub_out, sub_err = proc.communicate()
        code = proc.returncode
        if code != 0:
            raise Exception("gunzip failed for Postgres: %s, %s", code, sub_err)


class PgDump(IDump):
    """
    https://github.com/IBM/backwork-backup-postgresql/blob/master/postgresql/postgresql.py
    """
    _credentials: dict

    def __init__(self, conninfo: str):
        self._credentials = conninfo_to_dict(conninfo)

    async def dump(self, out: typing.IO[bytes]):
        os.environ["PGPASSWORD"] = self._credentials['password']
        cmd = [
            "pg_dump",
            "-c",
            "-h", self._credentials['host'],
            "-p", self._credentials['port'],
            "-U", self._credentials['user'],
            "-d", self._credentials['dbname'],
            "--no-password",
            # "--inserts",
            "--no-owner",
            "--no-privileges",
        ]
        proc = subprocess.Popen(cmd, env={'PGPASSWORD': self._credentials['password']},
                                stdout=out, stderr=subprocess.PIPE)
        sub_out, sub_err = proc.communicate()
        code = proc.returncode
        if code != 0:
            raise Exception("pg_dump failed for Postgres: %s, %s", proc.returncode, sub_err)

    async def load(self, in_: typing.IO[bytes]):
        os.environ["PGPASSWORD"] = self._credentials['password']
        cmd = [
            "psql",
            "-h", self._credentials['host'],
            "-p", self._credentials['port'],
            "-U", self._credentials['user'],
            "-d", self._credentials['dbname'],
            "--no-password",
        ]
        proc = subprocess.Popen(cmd, env={'PGPASSWORD': self._credentials['password']},
                                stdin=in_, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        sub_out, sub_err = proc.communicate()
        code = proc.returncode
        if code != 0:
            raise Exception("psql failed for Postgres: %s, %s", proc.returncode, sub_err)



import typing
from abc import ABCMeta

from ascetic_ddd.seedwork.domain.identity.interfaces import IAccessible
from ascetic_ddd.faker.domain.providers._mixins import BaseCompositeProvider, ObservableMixin
from ascetic_ddd.faker.domain.specification.interfaces import ISpecification
from ascetic_ddd.faker.domain.providers.interfaces import IEntityProvider
from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.faker.domain.values.empty import empty
from ascetic_ddd.observable.interfaces import IObservable


__all__ = ('IAggregateRepository', 'AggregateProvider',)

T_Input = typing.TypeVar("T_Input")
T_Output = typing.TypeVar("T_Output")


class IAggregateRepository(IObservable, typing.Protocol[T_Output]):

    async def insert(self, session: ISession, agg: T_Output):
        ...

    async def get(self, session: ISession, id_: IAccessible[typing.Any]) -> T_Output | None:
        ...

    async def update(self, session: ISession, agg: T_Output):
        ...

    async def find(self, session: ISession, specification: ISpecification) -> typing.Iterable[T_Output]:
        ...

    async def setup(self, session: ISession):
        ...

    async def cleanup(self, session: ISession):
        ...


class AggregateProvider(
    BaseCompositeProvider[T_Input, T_Output],
    IEntityProvider[T_Input, T_Output],
    typing.Generic[T_Input, T_Output],
    metaclass=ABCMeta
):
    _id_attr: str
    _repository: IAggregateRepository[T_Output]
    _result_factory: typing.Callable[[...], T_Output] = None  # T_Output of each nested Provider.
    _result_exporter: typing.Callable[[T_Output], T_Input] = None

    _aspect_mapping = {
        "repository": "_repository",
        **ObservableMixin._aspect_mapping
    }

    def __init__(
            self,
            repository: IAggregateRepository,
            # distributor_factory: IM2ODistributorFactory,
            result_factory: typing.Callable[[...], T_Output] | None = None,  # T_Output of each nested Provider.
            result_exporter: typing.Callable[[T_Output], T_Input] | None = None,
    ):
        self._repository = repository

        if self._result_factory is None:
            if result_factory is None:

                def result_factory(**result):
                    return result

            self._result_factory = result_factory

        if self._result_exporter is None:
            if result_exporter is None:

                def result_exporter(value):
                    return value

            self._result_exporter = result_exporter

        super().__init__()
        self.on_init()

    def on_init(self):
        pass

    async def create(self, session: ISession) -> T_Output:
        if self._output_result is not empty:
            return self._output_result
        result = await self._default_factory(session)
        if self.id_provider.is_complete() and not self.id_provider.is_transient():
            # id_ здесь может быть еще неизвестен, т.к. агрегат не создан.
            # А может быть и известен, если его id_ реиспользуется как FK.
            id_ = await self.id_provider.create(session)
            # Skip repository lookup if id contains empty fields (auto-increment PKs)
            saved_result = await self._repository.get(session, id_)
        else:
            saved_result = None

        if saved_result is not None:
            result = saved_result
            state = self._result_exporter(result)
            self.set(state)
            await self.populate(session)
        else:
            await self._repository.insert(session, result)
            state = self._result_exporter(result)
            self.id_provider.set(state.get(self._id_attr))
            await self.id_provider.populate(session)
            # await self.id_provider.append(session, getattr(result, self._id_attr))
        # self.set() could reset self._output_result
        self._output_result = result

        # Create dependent entities AFTER this aggregate is created (they need its ID for FK)
        if self._dependent_providers:
            dependency_id = self.id_provider.get()
            for attr, dep_provider in self._dependent_providers.items():
                dep_provider.set_dependency_id(dependency_id)
                await dep_provider.populate(session)
                await dep_provider.create(session)

        return result

    async def populate(self, session: ISession) -> None:
        # Prevent diamond problem (cycles in FK)
        # See also https://github.com/mikeboers/C3Linearize
        if self.is_complete():
            return
        await self.do_populate(session)
        for attr, provider in self._providers.items():
            await provider.populate(session)

    async def do_populate(self, session: ISession) -> None:
        pass

    async def _default_factory(self, session: ISession, position: typing.Optional[int] = None):
        data = dict()
        for attr, provider in self._providers.items():
            data[attr] = await provider.create(session)
        return self._result_factory(**data)

    @property
    def id_provider(self):
        return getattr(self, self._id_attr)

    async def setup(self, session: ISession):
        await self._repository.setup(session)
        await super().setup(session)

    async def cleanup(self, session: ISession):
        await self._repository.cleanup(session)
        await super().cleanup(session)



import typing
from enum import Enum


class Currency(str, Enum):
    """Валюта"""

    USD = "USD"
    EUR = "EUR"
    RUB = "RUB"
    GBP = "GBP"

    def to_symbol(self) -> str:
        """Получить символ валюты"""
        mapping = {
            Currency.USD: "$",
            Currency.EUR: "€",
            Currency.RUB: "₽",
            Currency.GBP: "£",
        }
        return mapping[self]

    def export(self, setter: typing.Callable[[str], None]) -> None:
        setter(self.value)



"""
Extension for jsonpath2 parser to support parameterized expressions.

Extends the parser to recognize placeholders (%s, %d, %f, %(name)s)
and bind values at execution time.
"""
from typing import Any, Dict, Tuple, Union, Iterator
import re
from jsonpath2.path import Path
from jsonpath2.node import MatchData
from jsonpath2.expression import Expression


class PlaceholderReference:
    """
    Reference to a placeholder location in the AST.

    Stores where a placeholder is located so we can update it when binding parameters.
    """

    def __init__(self, expression, attribute: str, name: str, format_type: str):
        """
        Initialize placeholder reference.

        Args:
            expression: The BinaryOperatorExpression that contains the placeholder
            attribute: The attribute name ('left_node_or_value' or 'right_node_or_value')
            name: Placeholder name (for named) or index (for positional)
            format_type: Type hint ('s', 'd', 'f')
        """
        self.expression = expression
        self.attribute = attribute
        self.name = name
        self.format_type = format_type

    def bind(self, value: Any):
        """Bind a value to this placeholder by updating the AST."""
        setattr(self.expression, self.attribute, value)


class ParametrizedPath:
    """
    JSONPath with placeholder support.

    Parses template once, binds different values at execution time.
    """

    def __init__(self, template: str):
        """
        Parse template with placeholders.

        Args:
            template: JSONPath with %s, %d, %f or %(name)s placeholders
        """
        self.template = template
        self._placeholder_info = []  # Info from preprocessing
        self._placeholder_refs = []  # References to AST locations
        self._placeholder_index = 0  # Track which placeholder to use next

        # Parse template and replace placeholders with markers
        processed_template, self._placeholder_info = self._preprocess_template(template)

        # Parse with jsonpath2
        self._path = Path.parse_str(processed_template)

        # Post-process AST to create placeholder references
        self._inject_placeholders()

    @property
    def placeholders(self):
        """Return placeholder info for compatibility."""
        return self._placeholder_info

    @staticmethod
    def _normalize_equality_operator(template: str) -> str:
        """
        Normalize == to = for jsonpath2 library compatibility.

        RFC 9535 standard defines == for equality, but jsonpath2 library
        deviates from the standard and uses single =.
        This method provides better UX by accepting both syntaxes.

        Args:
            template: JSONPath template string

        Returns:
            Normalized template with == replaced by =
        """
        result = []
        i = 0
        in_string = False
        string_char = None

        while i < len(template):
            char = template[i]

            # Track if we're inside a string literal
            if char in ('"', "'") and (i == 0 or template[i-1] != '\\'):
                if not in_string:
                    in_string = True
                    string_char = char
                elif char == string_char:
                    in_string = False
                    string_char = None

            # Replace == with = only outside strings
            if not in_string and char == '=' and i + 1 < len(template) and template[i + 1] == '=':
                result.append('=')
                i += 2  # Skip both = characters
                continue

            result.append(char)
            i += 1

        return ''.join(result)

    @staticmethod
    def _normalize_logical_operators(template: str) -> str:
        """
        Normalize RFC 9535 logical operators to jsonpath2 text operators.

        RFC 9535 standard defines: &&, ||, !
        jsonpath2 library uses text operators: and, or, not
        This method normalizes symbol operators to text operators.

        Args:
            template: JSONPath template string

        Returns:
            Normalized template with text logical operators
        """
        result = []
        i = 0
        in_string = False
        string_char = None

        while i < len(template):
            char = template[i]

            # Track if we're inside a string literal
            if char in ('"', "'") and (i == 0 or template[i-1] != '\\'):
                if not in_string:
                    in_string = True
                    string_char = char
                elif char == string_char:
                    in_string = False
                    string_char = None

            if not in_string:
                # Replace && with and
                if char == '&' and i + 1 < len(template) and template[i + 1] == '&':
                    result.append(' and ')
                    i += 2
                    continue

                # Replace || with or
                elif char == '|' and i + 1 < len(template) and template[i + 1] == '|':
                    result.append(' or ')
                    i += 2
                    continue

                # Replace ! with not (but not in !=)
                elif char == '!' and i + 1 < len(template) and template[i + 1] != '=':
                    # Special case: ?!(...) should become ?(not ...) not ?not (...)
                    # Check if previous non-space char is ?
                    if result and ''.join(result).rstrip().endswith('?'):
                        # Convert ?!(...) to ?(not ...)
                        # Skip the ! and the opening paren, add (not
                        if i + 1 < len(template) and template[i + 1] == '(':
                            result.append('(not ')
                            i += 2  # Skip ! and (
                            continue
                    result.append('not ')
                    i += 1
                    continue

            result.append(char)
            i += 1

        return ''.join(result)

    @staticmethod
    def _add_parentheses_to_filter(template: str) -> str:
        """
        Add parentheses around filter expressions if not present.

        jsonpath2 library requires parentheses: $[?(@.age > 25)] not $[?@.age > 25]
        RFC 9535 allows both syntaxes, so we normalize to jsonpath2 format.

        Args:
            template: JSONPath template string

        Returns:
            Template with parentheses added
        """
        result = template

        # Find all [?@ patterns that don't have ( after ?
        pattern = r'\[\?(?!\()'  # [? not followed by (
        positions = []
        for match in re.finditer(pattern, result):
            # Check if next char is @ or space then @
            pos = match.end()
            if pos < len(result) and (result[pos] == '@' or (result[pos] == ' ' and pos + 1 < len(result) and result[pos + 1] == '@')):
                positions.append(match.start())

        # Process from right to left to maintain positions
        for pos in reversed(positions):
            # Find the matching ]
            depth = 1
            i = pos + 2  # After [?
            closing_pos = None

            while i < len(result) and depth > 0:
                if result[i] == '[':
                    depth += 1
                elif result[i] == ']':
                    depth -= 1
                    if depth == 0:
                        closing_pos = i
                        break
                i += 1

            if closing_pos is not None:
                # Insert ) before ]
                result = result[:closing_pos] + ')' + result[closing_pos:]
                # Insert ( after ?
                insert_pos = pos + 2  # After [?
                result = result[:insert_pos] + '(' + result[insert_pos:]

        return result

    def _preprocess_template(self, template: str) -> Tuple[str, list]:
        """
        Replace placeholders with temporary markers and normalize operators.

        Returns:
            (processed_template, list of placeholder info)
        """
        placeholders = []

        # Add parentheses to filter expressions (required by jsonpath2)
        processed = self._add_parentheses_to_filter(template)

        # Normalize == to = for jsonpath2 library compatibility
        processed = self._normalize_equality_operator(processed)

        # Normalize logical operators for jsonpath2 library compatibility
        processed = self._normalize_logical_operators(processed)

        # Find named placeholders: %(name)s, %(age)d, %(price)f
        named_pattern = r'%\((\w+)\)([sdf])'
        for match in re.finditer(named_pattern, processed):
            name = match.group(1)
            format_type = match.group(2)
            placeholder_str = match.group(0)

            # Replace with a valid literal based on type
            if format_type == 's':
                replacement = '"__PLACEHOLDER__"'
            elif format_type == 'd':
                replacement = '999999'
            elif format_type == 'f':
                replacement = '999999.0'

            placeholders.append({
                'original': placeholder_str,
                'name': name,
                'format_type': format_type,
                'replacement': replacement,
                'positional': False
            })

            processed = processed.replace(placeholder_str, replacement, 1)

        # Find positional placeholders: %s, %d, %f
        positional_pattern = r'%([sdf])'
        position = 0
        for match in re.finditer(positional_pattern, processed):
            format_type = match.group(1)
            placeholder_str = match.group(0)

            # Replace with a valid literal
            if format_type == 's':
                replacement = '"__PLACEHOLDER__"'
            elif format_type == 'd':
                replacement = '999999'
            elif format_type == 'f':
                replacement = '999999.0'

            placeholders.append({
                'original': placeholder_str,
                'name': str(position),
                'format_type': format_type,
                'replacement': replacement,
                'positional': True
            })

            processed = processed.replace(placeholder_str, replacement, 1)
            position += 1

        return processed, placeholders

    def _inject_placeholders(self):
        """
        Walk the parsed AST and replace marker literals with PlaceholderExpression.

        This modifies the AST in-place.
        """
        # Traverse the node chain starting from root
        current_node = self._path.root_node
        while current_node:
            # Check if it's a SubscriptNode with subscripts
            if hasattr(current_node, 'subscripts'):
                for subscript in current_node.subscripts:
                    self._process_subscript(subscript)

            # Move to next node in chain
            current_node = getattr(current_node, 'next_node', None)

    def _process_subscript(self, subscript):
        """Process a subscript node, recursively replacing placeholders."""
        # Check if it's a FilterSubscript
        if hasattr(subscript, 'expression'):
            self._process_expression(subscript.expression)

    def _process_expression(self, expression):
        """Recursively process expression nodes."""
        # Handle VariadicOperatorExpression (AND/OR) which has 'expressions' list
        if hasattr(expression, 'expressions'):
            for sub_expr in expression.expressions:
                self._process_expression(sub_expr)
            return

        # Handle UnaryOperatorExpression (NOT) which has 'expression' in jsonpath2
        if hasattr(expression, 'expression') and not hasattr(expression, 'expressions'):
            inner = expression.expression
            if hasattr(inner, 'evaluate'):
                self._process_expression(inner)
            return

        # Check for BinaryOperatorExpression with left_node_or_value/right_node_or_value
        if hasattr(expression, 'left_node_or_value') and hasattr(expression, 'right_node_or_value'):
            # Check left and right for placeholder markers
            self._check_and_create_placeholder_ref(
                expression, 'left_node_or_value', expression.left_node_or_value
            )
            self._check_and_create_placeholder_ref(
                expression, 'right_node_or_value', expression.right_node_or_value
            )

            # Recursively process sub-expressions (if they have evaluate methods)
            if hasattr(expression.left_node_or_value, 'evaluate'):
                self._process_expression(expression.left_node_or_value)
            if hasattr(expression.right_node_or_value, 'evaluate'):
                self._process_expression(expression.right_node_or_value)

    def _check_and_create_placeholder_ref(self, expression, attribute: str, node_or_value):
        """Check if node_or_value is a placeholder marker and create reference."""
        # Check if this is one of our placeholder markers
        if self._placeholder_index < len(self._placeholder_info):
            ph = self._placeholder_info[self._placeholder_index]

            # Check if the value matches the expected marker
            is_match = False
            if ph['format_type'] == 's' and node_or_value == "__PLACEHOLDER__":
                is_match = True
            elif ph['format_type'] in ('d', 'f') and node_or_value == 999999:
                is_match = True

            if is_match:
                # Create a reference to this placeholder location
                placeholder_ref = PlaceholderReference(
                    expression, attribute, ph['name'], ph['format_type']
                )
                self._placeholder_refs.append(placeholder_ref)
                self._placeholder_index += 1

    def match(self, data: Any, params: Union[Tuple[Any, ...], Dict[str, Any]]) -> Iterator[MatchData]:
        """
        Match data with bound parameters.

        Args:
            data: Data to match
            params: Parameter values (tuple for positional, dict for named)

        Returns:
            Iterator of MatchData
        """
        # Bind parameter values to PlaceholderExpression nodes
        self._bind_placeholders(params)

        # Execute the path with bound parameters
        return self._path.match(data)

    def _bind_placeholders(self, params: Union[Tuple[Any, ...], Dict[str, Any]]):
        """Bind parameter values to all placeholders in AST."""
        for placeholder_ref in self._placeholder_refs:
            # Get the value for this placeholder
            if placeholder_ref.name.isdigit():
                # Positional
                idx = int(placeholder_ref.name)
                if idx >= len(params):
                    raise ValueError(f"Missing positional parameter at index {idx}")
                value = params[idx]
            else:
                # Named
                if placeholder_ref.name not in params:
                    raise ValueError(f"Missing named parameter: {placeholder_ref.name}")
                value = params[placeholder_ref.name]

            # Bind the value by updating the AST
            placeholder_ref.bind(value)

    def find(self, data: Any, params: Union[Tuple[Any, ...], Dict[str, Any]]) -> list:
        """Find all matching values."""
        return [match.current_value for match in self.match(data, params)]

    def find_one(self, data: Any, params: Union[Tuple[Any, ...], Dict[str, Any]]) -> Any:
        """Find first matching value."""
        for match in self.match(data, params):
            return match.current_value
        return None


def parse(template: str) -> ParametrizedPath:
    """
    Parse JSONPath expression with C-style placeholders.

    Args:
        template: JSONPath with %s, %d, %f or %(name)s placeholders

    Returns:
        ParametrizedPath that can be executed with different parameter values

    Examples:
        >>> path = parse("$[?(@.age > %d)]")
        >>> results = path.match(data, (27,))
        >>>
        >>> path = parse("$[?(@.name = %(name)s)]")
        >>> results = path.match(data, {"name": "Alice"})
    """
    return ParametrizedPath(template)



import typing

from ascetic_ddd.faker.domain.distributors.m2o.interfaces import ICursor

from ascetic_ddd.faker.domain.session.interfaces import ISession

__all__ = ("Cursor",)

T = typing.TypeVar("T", covariant=True)


class Cursor(ICursor, typing.Generic[T]):
    position: int | None
    _callback: typing.Callable[[ISession, T, int | None], typing.Awaitable[None]]

    def __init__(
            self, position: int | None,
            callback: typing.Callable[[ISession, T, int | None], typing.Awaitable[None]]
    ):
        self.position = position
        self._callback = callback

    async def append(self, session: ISession, value: T):
        await self._callback(session, value, self.position)



"""Composite expression node for handling composite keys."""
from typing import Protocol

from ..domain.nodes import (
    Visitable,
    Visitor,
    And,
    Equal,
    Not,
    NotEqual,
)


class CompositeExpressionsDifferentLengthError(Exception):
    """Raised when composite expressions have different lengths."""

    pass


class ICompositeExpression(Protocol):
    """Interface for expression composers."""

    def __eq__(self, other: "CompositeExpression") -> Visitable:
        """Create equality expression with another composite."""
        ...

    def __ne__(self, other: "CompositeExpression") -> Visitable:
        """Create not-equal expression with another composite."""
        ...

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        ...


class CompositeExpression(Visitable):
    """Node representing a composite expression (e.g., composite key)."""

    def __init__(self, *nodes: Visitable):
        self._nodes = list(nodes)

    def __eq__(self, other: "CompositeExpression") -> Visitable:
        """
        Create an AND expression of equality comparisons.

        For composite keys: (a1 = b1) AND (a2 = b2) AND ...
        """
        if len(self._nodes) != len(other._nodes):
            raise CompositeExpressionsDifferentLengthError(
                "Composite expressions have different length"
            )

        operands = []
        for i in range(len(self._nodes)):
            left, right = self._nodes[i], other._nodes[i]

            if isinstance(left, CompositeExpression):
                if not isinstance(right, CompositeExpression):
                    raise CompositeExpressionsDifferentLengthError(
                        "Composite expressions have different length"
                    )
                new_node = left == right
                operands.append(new_node)
            else:
                operands.append(Equal(left, right))

        return And(operands[0], *operands[1:])

    def __ne__(self, other: "CompositeExpression") -> Visitable:
        """
        Create a NOT(AND(...)) expression for inequality.

        For composite keys: NOT((a1 = b1) AND (a2 = b2) AND ...)
        """
        if len(self._nodes) != len(other._nodes):
            raise CompositeExpressionsDifferentLengthError(
                "Composite expressions have different length"
            )

        operands = []
        for i in range(len(self._nodes)):
            left, right = self._nodes[i], other._nodes[i]

            if isinstance(left, CompositeExpression):
                if not isinstance(right, CompositeExpression):
                    raise CompositeExpressionsDifferentLengthError(
                        "Composite expressions have different length"
                    )
                new_node = left != right
                operands.append(new_node)
            else:
                operands.append(NotEqual(left, right))

        return Not(And(operands[0], *operands[1:]))

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor (no-op for composite nodes)."""
        pass



from .event_bus import IEventBus, InMemoryEventBus
from .utils.amemo import amemo

__all__ = (
    "ascetic_ddd_factory",
    "BuildingBlocksFactory",
)


class BuildingBlocksFactory:
    @amemo
    async def make_in_memory_event_bus(self) -> IEventBus[str, dict]:
        return InMemoryEventBus[str, dict]()


ascetic_ddd_factory = BuildingBlocksFactory()



import datetime
import typing
import uuid

from .causal_dependency import CausalDependency
from .causal_dependency_exporter import CausalDependencyExporter
from .event_meta import IEventMetaExporter

__all__ = ("EventMetaExporter",)


class EventMetaExporter(IEventMetaExporter):
    def __init__(self) -> None:
        self.data: dict[str, typing.Any] = {
            "causal_dependencies": [],
        }

    def set_event_id(self, value: uuid.UUID | None) -> None:
        self.data["event_id"] = value

    def set_causation_id(self, value: uuid.UUID | None) -> None:
        self.data["causation_id"] = value

    def set_correlation_id(self, value: uuid.UUID | None) -> None:
        self.data["correlation_id"] = value

    def set_reason(self, value: str | None) -> None:
        self.data["reason"] = value

    def set_occurred_at(self, value: datetime.datetime | None) -> None:
        self.data["occurred_at"] = value

    def add_causal_dependency(self, value: CausalDependency) -> None:
        ex = CausalDependencyExporter()
        value.export(ex)
        self.data["causal_dependencies"].append(ex.data)



import functools
import json
import math
import os
import random
import socket
import threading
import typing
import dataclasses
from abc import abstractmethod

from psycopg.types.json import Jsonb

from ascetic_ddd.faker.domain.distributors.m2o.cursor import Cursor
from ascetic_ddd.faker.domain.distributors.m2o.interfaces import IM2ODistributor
from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.faker.domain.specification.empty_specification import EmptySpecification
from ascetic_ddd.faker.domain.specification.interfaces import ISpecification
from ascetic_ddd.faker.infrastructure.distributors.m2o.interfaces import IPgExternalSource
from ascetic_ddd.faker.infrastructure.session.pg_session import extract_internal_connection
from ascetic_ddd.faker.infrastructure.specification.pg_specification_visitor import PgSpecificationVisitor
from ascetic_ddd.faker.infrastructure.utils.json import JSONEncoder
from ascetic_ddd.observable.observable import Observable
from ascetic_ddd.seedwork.infrastructure.utils import serializer
from ascetic_ddd.seedwork.infrastructure.utils.pg import escape


__all__ = ('BasePgDistributor', 'PgWeightedDistributor')


T = typing.TypeVar("T", covariant=True)


class BasePgDistributor(Observable, IM2ODistributor[T], typing.Generic[T]):
    """
    Базовый класс для PostgreSQL дистрибьюторов.

    Ограничение: при динамическом создании значений ранние значения
    получают больше вызовов, т.к. доступны дольше. Для генератора фейковых данных приемлемо.
    """
    _extract_connection = staticmethod(extract_internal_connection)
    _initialized: bool = False
    _mean: float = 50
    _values_table: str | None = None
    _default_key: str = str(frozenset())
    _provider_name: str | None = None
    _external_source: IPgExternalSource | None = None

    def __init__(
            self,
            mean: float | None = None,
            initialized: bool = False
    ):
        self._external_source = None
        if mean is not None:
            self._mean = mean
        self._initialized = initialized
        super().__init__()

    def bind_external_source(self, external_source: typing.Any) -> None:
        """Привязывает внешний источник данных (repository) и обновляет таблицу."""
        if not isinstance(external_source, IPgExternalSource):
            raise TypeError("Expected IPgExternalSource, got %s" % type(external_source))
        self._external_source = external_source
        self._values_table = external_source.table

    async def next(
            self,
            session: ISession,
            specification: ISpecification[T] | None = None,
    ) -> T:
        if specification is None:
            specification = EmptySpecification()

        if not self._initialized:
            await self.setup(session)

        if self._mean == 1:
            raise Cursor(
                position=None,
                callback=self._append,
            )

        # Резолвим вложенные constraints (если есть)
        if hasattr(specification, 'resolve_nested'):
            await specification.resolve_nested(session)

        value, should_create_new = await self._get_next_value(session, specification)
        if should_create_new:
            raise Cursor(
                position=None,
                callback=self._append,
            )
        return value

    @abstractmethod
    async def _get_next_value(self, session: ISession, specification: ISpecification[T]) -> tuple[T | None, bool]:
        raise NotImplementedError

    async def _append(self, session: ISession, value: T, position: int | None):
        if self._external_source:
            return
        sql = """
            INSERT INTO %(values_table)s (value, object)
            VALUES (%%s, %%s)
            ON CONFLICT DO NOTHING;
        """ % {
            'values_table': self._values_table,
        }
        async with self._extract_connection(session).cursor() as acursor:
            await acursor.execute(sql, (self._encode(value), self._serialize(value)))
        # logging.debug("Append: %s", value)
        await self.anotify('value', session, value)

    async def append(self, session: ISession, value: T):
        await self._append(session, value, None)

    @property
    def provider_name(self):
        return self._provider_name

    @provider_name.setter
    def provider_name(self, value):
        if self._provider_name is None:
            self._provider_name = value
            if self._values_table is None:
                self._values_table = escape("values_for_%s" % value[-(63 - 11):])

    async def setup(self, session: ISession):
        if not self._initialized:  # Fixes diamond problem
            if not (await self._is_initialized(session)):
                await self._setup(session)
            self._initialized = True

    async def cleanup(self, session: ISession):
        # FIXME: diamond problem
        self._initialized = False
        if self._external_source:
            return
        async with self._extract_connection(session).cursor() as acursor:
            await acursor.execute("DROP TABLE IF EXISTS %s" % self._values_table)

    def __copy__(self):
        return self

    def __deepcopy__(self, memodict={}):
        return self

    async def _setup(self, session: ISession):
        if self._external_source:
            return
        sql = """
            CREATE TABLE IF NOT EXISTS %(values_table)s (
                position serial NOT NULL PRIMARY KEY,
                value JSONB NOT NULL,
                object TEXT NOT NULL,
                UNIQUE (value)
            );
            CREATE INDEX IF NOT EXISTS %(index_name)s ON %(values_table)s USING GIN(value jsonb_path_ops);
        """ % {
            "values_table": self._values_table,
            "index_name": escape("gin_%s" % self.provider_name[:(63 - 4)]),
        }
        async with self._extract_connection(session).cursor() as acursor:
            await acursor.execute(sql)

    async def _is_initialized(self, session: ISession) -> bool:
        if self._external_source:
            return True  # external source table is managed by repository
        sql = """SELECT to_regclass(%s)"""
        async with self._extract_connection(session).cursor() as acursor:
            await acursor.execute(sql, (self._values_table,))
            regclass = (await acursor.fetchone())[0]
        return regclass is not None

    @staticmethod
    def get_thread_id():
        return '{0}.{1}.{2}'.format(
            socket.gethostname(), os.getpid(), threading.get_ident()
        )

    @staticmethod
    def _encode(obj):
        if dataclasses.is_dataclass(obj):
            obj = dataclasses.asdict(obj)
        dumps = functools.partial(json.dumps, cls=JSONEncoder)
        return Jsonb(obj, dumps)

    _serialize = staticmethod(serializer.serialize)
    _deserialize = staticmethod(serializer.deserialize)


# =============================================================================
# PgWeightedDistributor
# =============================================================================

class PgWeightedDistributor(BasePgDistributor[T], typing.Generic[T]):
    """
    Дистрибьютор с взвешенным распределением в PostgreSQL.

    Ограничение: при динамическом создании значений ранние значения
    получают больше вызовов, т.к. доступны дольше. Это даёт ~85% vs 70% для первой
    партиции вместо точного соответствия весам. Для генератора фейковых данных приемлемо.
    """
    _weights: list[float]

    def __init__(
            self,
            weights: typing.Iterable[float] = tuple(),
            mean: float | None = None,
            initialized: bool = False
    ):
        self._weights = list(weights)
        super().__init__(mean=mean, initialized=initialized)

    def _compute_partition(self) -> tuple[int, float, int]:
        """
        Вычисляет партицию в Python.

        Returns:
            (partition_idx, local_skew, num_partitions)

        Используем ЛЕВУЮ партицию (LAG) и смещаем к КОНЦУ — это компенсирует то, что ранние
        значения получают больше вызовов (доступны дольше при динамическом создании).
        Для weights=[0.7, 0.2, 0.07, 0.03]:
          partition 0: первая → local_skew=1.0 (равномерно)
          partition 1: ratio=3.5 → local_skew≈2.81 (смещение к концу, ближе к partition 0)
          partition 2: ratio=2.86 → local_skew≈2.52
          partition 3: ratio=2.33 → local_skew≈2.22
        """
        num_partitions = len(self._weights)
        if num_partitions == 0:
            return (0, 1.0, 1)

        # Выбор партиции по весам — O(w)
        partition_idx = random.choices(range(num_partitions), weights=self._weights, k=1)[0]

        # Вычисляем локальный наклон из соотношения весов соседних партиций
        if partition_idx > 0:
            prev_weight = self._weights[partition_idx - 1]
            curr_weight = self._weights[partition_idx]
            if curr_weight > 0:
                ratio = prev_weight / curr_weight
                local_skew = max(1.0, math.log2(ratio) + 1)
            else:
                local_skew = 2.0
        else:
            local_skew = 1.0

        return (partition_idx, local_skew, num_partitions)

    async def _get_next_value(self, session: ISession, specification: ISpecification[T]) -> tuple[T | None, bool]:
        """
        Оптимизированный выбор значения:
        1. Выбор партиции по весам — O(w) в Python
        2. Выбор позиции внутри партиции со slope bias — O(1) в SQL
        3. Получение значения по позиции — O(log n) с индексом
        4. Вероятностное решение о создании нового значения
        """
        visitor = PgSpecificationVisitor()
        specification.accept(visitor)

        partition_idx, local_skew, num_partitions = self._compute_partition()

        sql = """
            WITH filtered AS (
                SELECT position, object FROM %(values_table)s %(where)s
            ),
            stats AS (
                SELECT COUNT(*) AS n FROM filtered
            ),
            target AS (
                SELECT
                    -- end = floor((partition_idx + 1) * total / num_partitions)
                    -- size = ceil(total / num_partitions)
                    -- pos = end - 1 - floor(size * (1 - random())^local_skew)
                    -- Смещение к КОНЦУ партиции (ближе к предыдущей)
                    GREATEST(0,
                        FLOOR((%(partition_idx)s + 1) * n::decimal / %(num_partitions)s)::integer - 1 -
                        LEAST(
                            FLOOR(CEIL(n::decimal / %(num_partitions)s) * POWER(1 - RANDOM(), %(local_skew)s))::integer,
                            GREATEST(CEIL(n::decimal / %(num_partitions)s)::integer - 1, 0)
                        )
                    ) AS pos,
                    n
                FROM stats
            )
            SELECT
                (SELECT object FROM filtered ORDER BY position OFFSET t.pos LIMIT 1),
                -- Вероятностный подход: создаём новое с вероятностью 1/mean
                (t.n = 0 OR RANDOM() < 1.0 / %(expected_mean)s),
                t.n
            FROM target t
        """ % {
            'values_table': self._values_table,
            'where': "WHERE %s" % visitor.sql if visitor.sql else "",
            'partition_idx': partition_idx,
            'num_partitions': num_partitions,
            'local_skew': local_skew,
            'expected_mean': max(self._mean, 1),
        }

        async with self._extract_connection(session).cursor() as acursor:
            await acursor.execute(sql, visitor.params)
            row = await acursor.fetchone()
            if not row or not row[0]:
                return (None, True)
            should_create_new = row[1] if row[2] and row[2] > 0 else True
            return (self._deserialize(row[0]), should_create_new)



import math
import random
import typing

from ascetic_ddd.faker.domain.distributors.o2m.interfaces import IO2MDistributor

__all__ = ('WeightedDistributor',)


class WeightedDistributor(IO2MDistributor):
    """
    O2M дистрибьютор с взвешенным распределением по партициям.

    Параметры:
    - weights: веса партиций (например [0.7, 0.2, 0.07, 0.03])
    - mean: среднее количество items на owner

    Пример: weights=[0.7, 0.2, 0.07, 0.03], mean=50
    - 25% вызовов попадут в партицию 0 (крупные) — получат больше mean
    - 25% вызовов попадут в партицию 3 (мелкие) — получат меньше mean
    - Среднее по всем вызовам = mean

    Пример:
        dist = WeightedDistributor(weights=[0.7, 0.2, 0.07, 0.03], mean=50)
        devices_count = dist.distribute()  # среднее = 50
    """
    _weights: list[float]
    _mean: float

    def __init__(
            self,
            weights: typing.Iterable[float] = tuple(),
            mean: float | None = None,
    ):
        self._weights = list(weights) if weights else [0.7, 0.2, 0.07, 0.03]
        if not self._weights:
            self._weights = [1.0]
        self._mean = mean if mean is not None else 50.0

    def distribute(self) -> int:
        """
        Возвращает количество items.

        Returns:
            Случайное количество items (из распределения Пуассона).
            Среднее по всем вызовам = mean.
        """
        # Выбираем случайную позицию в распределении [0, 1)
        position = random.random()

        # Вычисляем ожидаемое количество для этой позиции
        expected = self._compute_expected_for_position(position)

        if expected <= 0:
            return 0

        return self._poisson(expected)

    def _compute_expected_for_position(self, position: float) -> float:
        """Вычисляет ожидаемое количество items для позиции в распределении."""
        num_partitions = len(self._weights)

        # Нормализуем веса
        total_weight = sum(self._weights)
        if total_weight == 0:
            return self._mean

        # Размер партиции (доля от позиций)
        partition_size = 1.0 / num_partitions

        # Определяем партицию по позиции
        partition_idx = min(int(position / partition_size), num_partitions - 1)

        # Позиция внутри партиции [0, 1)
        local_position = (position - partition_idx * partition_size) / partition_size

        # Вес партиции (доля items)
        partition_weight = self._weights[partition_idx] / total_weight

        # Среднее для этой партиции = mean * partition_weight * num_partitions
        # (т.к. партиция содержит 1/num_partitions позиций, но получает partition_weight items)
        average_in_partition = self._mean * partition_weight * num_partitions

        # Local skew внутри партиции (как в M2O)
        if partition_idx > 0 and self._weights[partition_idx] > 0:
            ratio = self._weights[partition_idx - 1] / self._weights[partition_idx]
            local_skew = max(1.0, math.log2(ratio) + 1)
        else:
            local_skew = 1.0

        # При local_skew=1: равномерное распределение внутри партиции
        if local_skew <= 1.01:
            return average_in_partition

        # Степенное распределение внутри партиции
        individual_weight = (1 - local_position) ** local_skew
        average_weight = 1.0 / (local_skew + 1)

        return average_in_partition * individual_weight / average_weight

    @staticmethod
    def _poisson(lam: float) -> int:
        """Генерирует случайное число из распределения Пуассона."""
        if lam > 30:
            result = random.gauss(lam, lam ** 0.5)
            return max(0, round(result))

        L = 2.718281828 ** (-lam)
        k = 0
        p = 1.0
        while p > L:
            k += 1
            p *= random.random()
        return k - 1



import typing

from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.faker.domain.specification.interfaces import ISpecification
from ascetic_ddd.faker.infrastructure.distributors.m2o.pg_weighted_distributor import BasePgDistributor
from ascetic_ddd.faker.infrastructure.specification.pg_specification_visitor import PgSpecificationVisitor


__all__ = ('PgSkewDistributor',)


T = typing.TypeVar("T", covariant=True)


class PgSkewDistributor(BasePgDistributor[T], typing.Generic[T]):
    """
    Дистрибьютор со степенным распределением в PostgreSQL.

    Один параметр skew вместо списка весов:
    - skew = 1.0 — равномерное распределение
    - skew = 2.0 — умеренный перекос (первые 20% получают ~60% вызовов)
    - skew = 3.0 — сильный перекос (первые 10% получают ~70% вызовов)

    Преимущества перед PgDistributor:
    - Один параметр вместо списка весов
    - Проще SQL (нет таблицы weights, нет cumulative weights)
    - O(1) выбор позиции

    Ограничение: при динамическом создании значений ранние значения
    получают больше вызовов, т.к. доступны дольше. Для генератора фейковых данных приемлемо.
    """
    _skew: float = 2.0

    def __init__(
            self,
            skew: float = 2.0,
            mean: float | None = None,
            initialized: bool = False
    ):
        self._skew = skew
        super().__init__(mean=mean, initialized=initialized)

    async def _get_next_value(self, session: ISession, specification: ISpecification[T]) -> tuple[T | None, bool]:
        """
        Выбор значения со степенным распределением:
        idx = floor(total_values * (1 - random())^skew)

        При skew=1: равномерное распределение
        При skew=2: первые 50% получают ~75% вызовов
        При skew=3: первые 33% получают ~70% вызовов

        Вероятностный подход для создания новых значений: с вероятностью 1/mean.
        Работает корректно per-specification (WHERE условие учитывается).
        """
        visitor = PgSpecificationVisitor()
        specification.accept(visitor)

        sql = """
            WITH filtered AS (
                SELECT position, object FROM %(values_table)s %(where)s
            ),
            stats AS (
                SELECT COUNT(*) AS n FROM filtered
            ),
            target AS (
                SELECT
                    -- Степенное распределение: idx = floor(n * (1 - random())^skew)
                    -- skew=1: равномерное, skew=2+: перекос к началу
                    LEAST(FLOOR(n * POWER(1 - RANDOM(), %(skew)s))::integer, GREATEST(n - 1, 0)) AS pos,
                    n
                FROM stats
            )
            SELECT
                (SELECT object FROM filtered ORDER BY position OFFSET t.pos LIMIT 1),
                -- Вероятностный подход: создаём новое с вероятностью 1/mean
                (t.n = 0 OR RANDOM() < 1.0 / %(expected_mean)s),
                t.n
            FROM target t
        """ % {
            'values_table': self._values_table,
            'where': "WHERE %s" % visitor.sql if visitor.sql else "",
            'skew': self._skew,
            'expected_mean': max(self._mean, 1),
        }

        async with self._extract_connection(session).cursor() as acursor:
            await acursor.execute(sql, visitor.params)
            row = await acursor.fetchone()
            if not row or not row[0]:
                return (None, True)
            should_create_new = row[1] if row[2] and row[2] > 0 else True
            return (self._deserialize(row[0]), should_create_new)



"""
JSONPath to Raw SQL compiler using jsonpath2 library.

Compiles JSONPath expressions into raw SQL queries for normalized relational databases.
No ORM or Query Builder dependencies - pure SQL generation.
"""
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple
from enum import Enum

from jsonpath2.nodes.root import RootNode
from jsonpath2.nodes.subscript import SubscriptNode
from jsonpath2.nodes.terminal import TerminalNode
from jsonpath2.subscripts.objectindex import ObjectIndexSubscript
from jsonpath2.subscripts.filter import FilterSubscript
from jsonpath2.subscripts.wildcard import WildcardSubscript
from jsonpath2.expressions.operator import (
    BinaryOperatorExpression,
    EqualBinaryOperatorExpression,
    AndVariadicOperatorExpression,
    OrVariadicOperatorExpression,
)
from jsonpath2.path import Path


class RelationType(Enum):
    """Type of relationship between tables."""
    ONE_TO_MANY = "one-to-many"
    MANY_TO_ONE = "many-to-one"
    MANY_TO_MANY = "many-to-many"


@dataclass
class ColumnDef:
    """Column definition."""
    name: str
    type: str  # SQL type: INTEGER, VARCHAR, etc
    nullable: bool = True
    primary_key: bool = False


@dataclass
class TableDef:
    """Table definition."""
    name: str
    columns: Dict[str, ColumnDef]  # column_name -> ColumnDef
    primary_key: str | Tuple[str, ...] = "id"  # Support composite PK

    def has_column(self, column_name: str) -> bool:
        """Check if table has column."""
        return column_name in self.columns

    def get_column_list(self) -> List[str]:
        """Get list of column names."""
        return list(self.columns.keys())

    def get_primary_key_columns(self) -> List[str]:
        """Get primary key columns as list."""
        if isinstance(self.primary_key, str):
            return [self.primary_key]
        return list(self.primary_key)


@dataclass
class RelationshipDef:
    """Relationship definition."""
    target_table: str
    foreign_key: str | Tuple[str, ...]  # FK column(s) in source table, supports composite FK
    target_primary_key: str | Tuple[str, ...] = "id"  # Target PK column(s), supports composite PK
    relationship_type: RelationType = RelationType.ONE_TO_MANY

    def get_foreign_key_columns(self) -> List[str]:
        """Get foreign key columns as list."""
        if isinstance(self.foreign_key, str):
            return [self.foreign_key]
        return list(self.foreign_key)

    def get_target_primary_key_columns(self) -> List[str]:
        """Get target primary key columns as list."""
        if isinstance(self.target_primary_key, str):
            return [self.target_primary_key]
        return list(self.target_primary_key)


@dataclass
class SchemaDef:
    """Schema definition."""
    tables: Dict[str, TableDef]  # table_name -> TableDef
    relationships: Dict[str, Dict[str, RelationshipDef]]  # table -> {field -> relationship}
    root_table: str

    def get_table(self, table_name: str) -> TableDef:
        """Get table definition."""
        if table_name not in self.tables:
            raise ValueError(f"Table '{table_name}' not found in schema")
        return self.tables[table_name]


class SQLQuery:
    """Raw SQL query builder."""

    def __init__(self, root_table: str):
        """Initialize query with root table."""
        self.root_table = root_table
        self.select_columns: List[str] = []  # Qualified column names
        self.from_table = root_table
        self.joins: List[str] = []  # JOIN clauses
        self.where_conditions: List[str] = []  # WHERE conditions
        self.joined_tables: Set[str] = {root_table}

    def add_select_column(self, table: str, column: str):
        """Add column to SELECT."""
        qualified = f"{table}.{column}"
        if qualified not in self.select_columns:
            self.select_columns.append(qualified)

    def add_select_all(self, table: str):
        """Add all columns from table to SELECT."""
        qualified = f"{table}.*"
        if qualified not in self.select_columns:
            self.select_columns.append(qualified)

    def add_join(
        self,
        target_table: str,
        source_table: str,
        foreign_key: str | Tuple[str, ...],
        target_primary_key: str | Tuple[str, ...] = "id",
    ):
        """Add JOIN clause. Supports composite keys."""
        if target_table in self.joined_tables:
            return  # Already joined

        # Handle composite keys
        fk_columns = [foreign_key] if isinstance(foreign_key, str) else list(foreign_key)
        pk_columns = [target_primary_key] if isinstance(target_primary_key, str) else list(target_primary_key)

        if len(fk_columns) != len(pk_columns):
            raise ValueError(
                f"Foreign key and primary key must have same number of columns: "
                f"{len(fk_columns)} != {len(pk_columns)}"
            )

        # Build join conditions
        join_conditions = [
            f"{source_table}.{fk} = {target_table}.{pk}"
            for fk, pk in zip(fk_columns, pk_columns)
        ]

        join_clause = f"JOIN {target_table} ON {' AND '.join(join_conditions)}"
        self.joins.append(join_clause)
        self.joined_tables.add(target_table)

    def add_where(self, condition: str):
        """Add WHERE condition."""
        if condition:
            self.where_conditions.append(condition)

    def build(self) -> str:
        """Build final SQL query."""
        # SELECT clause
        if self.select_columns:
            select_clause = f"SELECT {', '.join(self.select_columns)}"
        else:
            select_clause = f"SELECT {self.from_table}.*"

        # FROM clause
        from_clause = f"FROM {self.from_table}"

        # JOIN clauses
        join_clause = "\n".join(self.joins) if self.joins else ""

        # WHERE clause
        where_clause = ""
        if self.where_conditions:
            where_clause = f"WHERE {' AND '.join(self.where_conditions)}"

        # Combine all parts
        parts = [select_clause, from_clause]
        if join_clause:
            parts.append(join_clause)
        if where_clause:
            parts.append(where_clause)

        return "\n".join(parts)


class CompilationContext:
    """Context for tracking compilation state."""

    def __init__(self, schema: SchemaDef):
        self.schema = schema
        self.current_table = schema.root_table
        self.query = SQLQuery(schema.root_table)

    def get_table(self, table_name: str) -> TableDef:
        """Get table definition."""
        return self.schema.get_table(table_name)

    def get_current_table(self) -> TableDef:
        """Get current table being processed."""
        return self.get_table(self.current_table)

    def get_relationship(self, field_name: str) -> Optional[RelationshipDef]:
        """Get relationship for field in current table."""
        if self.current_table not in self.schema.relationships:
            return None
        return self.schema.relationships[self.current_table].get(field_name)

    def add_join(self, target_table: str, relationship: RelationshipDef):
        """Add JOIN to target table. Supports composite keys and relationship types."""
        if target_table in self.query.joined_tables:
            return  # Already joined

        source_table = self.current_table

        # Handle composite keys
        fk = relationship.foreign_key
        pk = relationship.target_primary_key
        fk_columns = [fk] if isinstance(fk, str) else list(fk)
        pk_columns = [pk] if isinstance(pk, str) else list(pk)

        if len(fk_columns) != len(pk_columns):
            raise ValueError(
                f"Foreign key and primary key must have same number of columns: "
                f"{len(fk_columns)} != {len(pk_columns)}"
            )

        # Create join conditions based on relationship type
        # For ONE_TO_MANY: FK is in target table, PK is in source table
        # For MANY_TO_ONE: FK is in source table, PK is in target table
        if relationship.relationship_type == RelationType.ONE_TO_MANY:
            # FK columns in target table, PK columns in source table
            join_conds = [
                f"{source_table}.{pk} = {target_table}.{fk}"
                for fk, pk in zip(fk_columns, pk_columns)
            ]
        else:  # MANY_TO_ONE or MANY_TO_MANY
            # FK columns in source table, PK columns in target table
            join_conds = [
                f"{source_table}.{fk} = {target_table}.{pk}"
                for fk, pk in zip(fk_columns, pk_columns)
            ]

        join_condition = " AND ".join(join_conds)
        self.query.joins.append(f"JOIN {target_table} ON {join_condition}")
        self.query.joined_tables.add(target_table)

    def add_where(self, condition: str):
        """Add WHERE condition."""
        self.query.add_where(condition)

    def add_select_column(self, column: str):
        """Add column to SELECT."""
        self.query.add_select_column(self.current_table, column)

    def add_select_all(self):
        """Add all columns from current table to SELECT."""
        self.query.add_select_all(self.current_table)


class NodeVisitor(ABC):
    """Abstract base visitor for AST nodes."""

    @abstractmethod
    def visit(self, node, context: CompilationContext):
        """Visit node and compile to SQL."""
        pass


class RootNodeVisitor(NodeVisitor):
    """Visitor for RootNode ($)."""

    def visit(self, node: RootNode, context: CompilationContext):
        """Visit root node - initialize from root table."""
        visitor = get_visitor(node.next_node)
        return visitor.visit(node.next_node, context)


class SubscriptNodeVisitor(NodeVisitor):
    """Visitor for SubscriptNode ([...])."""

    def visit(self, node: SubscriptNode, context: CompilationContext):
        """Visit subscript node - process each subscript."""
        for subscript in node.subscripts:
            subscript_visitor = get_subscript_visitor(subscript)
            subscript_visitor.visit(subscript, context)

        # Continue to next node
        if not isinstance(node.next_node, TerminalNode):
            visitor = get_visitor(node.next_node)
            return visitor.visit(node.next_node, context)


class TerminalNodeVisitor(NodeVisitor):
    """Visitor for TerminalNode (end of path)."""

    def visit(self, node: TerminalNode, context: CompilationContext):
        """Visit terminal node - compilation complete."""
        pass


class SubscriptVisitor(ABC):
    """Abstract base visitor for subscript nodes."""

    @abstractmethod
    def visit(self, subscript, context: CompilationContext):
        """Visit subscript and update context."""
        pass


class ObjectIndexSubscriptVisitor(SubscriptVisitor):
    """Visitor for ObjectIndexSubscript (.field or ['field'])."""

    def visit(self, subscript: ObjectIndexSubscript, context: CompilationContext):
        """Visit object index subscript."""
        field_name = subscript.index

        # Check if this is a relationship
        relationship = context.get_relationship(field_name)

        if relationship:
            # This is a relationship - add JOIN
            context.add_join(relationship.target_table, relationship)
            context.current_table = relationship.target_table
        else:
            # This is a regular column - add to SELECT
            current_table = context.get_current_table()
            if current_table.has_column(field_name):
                context.add_select_column(field_name)
            else:
                raise ValueError(
                    f"Column '{field_name}' not found in table '{context.current_table}'"
                )


class FilterSubscriptVisitor(SubscriptVisitor):
    """Visitor for FilterSubscript ([?(expression)])."""

    def visit(self, subscript: FilterSubscript, context: CompilationContext):
        """Visit filter subscript - compile to WHERE clause."""
        expression = subscript.expression
        condition = self._compile_expression(expression, context)
        context.add_where(condition)

    def _compile_expression(self, expression, context: CompilationContext) -> str:
        """Compile any expression type to SQL condition."""
        if isinstance(expression, AndVariadicOperatorExpression):
            return self._compile_variadic_operator(expression, context, "AND")
        elif isinstance(expression, OrVariadicOperatorExpression):
            return self._compile_variadic_operator(expression, context, "OR")
        elif isinstance(expression, BinaryOperatorExpression):
            return self._compile_binary_operator(expression, context)
        else:
            raise NotImplementedError(f"Expression type {type(expression)} not supported")

    def _compile_variadic_operator(
        self, expr, context: CompilationContext, operator: str
    ) -> str:
        """Compile variadic operator expression (AND/OR) to SQL condition."""
        conditions = []
        for sub_expr in expr.expressions:
            condition = self._compile_expression(sub_expr, context)
            conditions.append(condition)
        return f"({f' {operator} '.join(conditions)})"

    def _compile_binary_operator(
        self, expr: BinaryOperatorExpression, context: CompilationContext
    ) -> str:
        """Compile binary operator expression to SQL condition."""
        # Get left operand
        if hasattr(expr.left_node_or_value, "__jsonpath__"):
            left_column = self._compile_node_to_column(expr.left_node_or_value, context)
        else:
            left_value = self._format_value(expr.left_node_or_value)
            left_column = left_value

        # Get right operand
        if hasattr(expr.right_node_or_value, "__jsonpath__"):
            right_column = self._compile_node_to_column(expr.right_node_or_value, context)
        else:
            right_value = self._format_value(expr.right_node_or_value)
            right_column = right_value

        # Compile operator
        operator_map = {
            "=": "=",
            "!=": "!=",
            ">": ">",
            "<": "<",
            ">=": ">=",
            "<=": "<=",
        }

        operator = operator_map.get(expr.token, expr.token)
        return f"{left_column} {operator} {right_column}"

    def _compile_node_to_column(self, node, context: CompilationContext) -> str:
        """Compile a node reference (like @.field or @.orders.total) to qualified column name."""
        from jsonpath2.nodes.current import CurrentNode
        from jsonpath2.nodes.terminal import TerminalNode

        if isinstance(node, CurrentNode):
            # Collect all field names in the path
            field_names = []
            current = node.next_node

            while current is not None:
                if isinstance(current, SubscriptNode):
                    for subscript in current.subscripts:
                        if isinstance(subscript, ObjectIndexSubscript):
                            field_names.append(subscript.index)
                    current = current.next_node
                elif isinstance(current, TerminalNode):
                    break
                else:
                    break

            if not field_names:
                raise NotImplementedError("Empty path in filter expression")

            # Simple path: @.field
            if len(field_names) == 1:
                return f"{context.current_table}.{field_names[0]}"

            # Nested path: @.orders.total - need to add JOINs
            return self._compile_nested_path(field_names, context)

        raise NotImplementedError(f"Node type {type(node)} not supported in filter")

    def _compile_nested_path(self, field_names: list, context: CompilationContext) -> str:
        """
        Compile nested path to SQL with JOINs.

        Args:
            field_names: List of field names in the path (e.g., ['orders', 'total'])
            context: Compilation context

        Returns:
            Qualified column reference (e.g., "orders.total")
        """
        # Save current table to restore later
        saved_table = context.current_table

        # Navigate through relationships (all fields except the last)
        for field_name in field_names[:-1]:
            relationship = context.get_relationship(field_name)

            if not relationship:
                raise ValueError(
                    f"Field '{field_name}' is not a relationship in table '{context.current_table}'"
                )

            # Add JOIN for this relationship
            context.add_join(relationship.target_table, relationship)
            context.current_table = relationship.target_table

        # Get the final field name
        final_field = field_names[-1]
        target_table = context.current_table

        # Verify the field exists in the target table
        table_def = context.schema.get_table(target_table)
        if not table_def.has_column(final_field):
            raise ValueError(
                f"Column '{final_field}' not found in table '{target_table}'"
            )

        # Restore original table context
        context.current_table = saved_table

        # Return qualified column reference
        return f"{target_table}.{final_field}"

    def _format_value(self, value) -> str:
        """Format Python value to SQL literal."""
        if value is None:
            return "NULL"
        elif isinstance(value, bool):
            return "TRUE" if value else "FALSE"
        elif isinstance(value, (int, float)):
            return str(value)
        elif isinstance(value, str):
            # Escape single quotes
            escaped = value.replace("'", "''")
            return f"'{escaped}'"
        else:
            raise ValueError(f"Unsupported value type: {type(value)}")


class WildcardSubscriptVisitor(SubscriptVisitor):
    """Visitor for WildcardSubscript ([*])."""

    def visit(self, subscript: WildcardSubscript, context: CompilationContext):
        """Visit wildcard subscript - select all from current table."""
        context.add_select_all()


# Visitor registry
_node_visitors = {
    RootNode: RootNodeVisitor(),
    SubscriptNode: SubscriptNodeVisitor(),
    TerminalNode: TerminalNodeVisitor(),
}

_subscript_visitors = {
    ObjectIndexSubscript: ObjectIndexSubscriptVisitor(),
    FilterSubscript: FilterSubscriptVisitor(),
    WildcardSubscript: WildcardSubscriptVisitor(),
}


def get_visitor(node) -> NodeVisitor:
    """Get visitor for node type."""
    node_type = type(node)
    if node_type not in _node_visitors:
        raise NotImplementedError(f"No visitor for node type {node_type}")
    return _node_visitors[node_type]


def get_subscript_visitor(subscript) -> SubscriptVisitor:
    """Get visitor for subscript type."""
    subscript_type = type(subscript)
    if subscript_type not in _subscript_visitors:
        raise NotImplementedError(f"No visitor for subscript type {subscript_type}")
    return _subscript_visitors[subscript_type]


class JSONPathToRawSQLCompiler:
    """Compiler for JSONPath expressions to raw SQL queries."""

    def __init__(self, schema: SchemaDef):
        """Initialize compiler with schema definition."""
        self.schema = schema

    @staticmethod
    def _normalize_equality_operator(template: str) -> str:
        """
        Normalize == to = for jsonpath2 library compatibility.

        RFC 9535 standard defines == for equality, but jsonpath2 library
        deviates from the standard and uses single =.
        This method provides better UX by accepting both syntaxes.

        Args:
            template: JSONPath template string

        Returns:
            Normalized template with == replaced by =
        """
        result = []
        i = 0
        in_string = False
        string_char = None

        while i < len(template):
            char = template[i]

            # Track if we're inside a string literal
            if char in ('"', "'") and (i == 0 or template[i-1] != '\\'):
                if not in_string:
                    in_string = True
                    string_char = char
                elif char == string_char:
                    in_string = False
                    string_char = None

            # Replace == with = only outside strings
            if not in_string and char == '=' and i + 1 < len(template) and template[i + 1] == '=':
                result.append('=')
                i += 2  # Skip both = characters
                continue

            result.append(char)
            i += 1

        return ''.join(result)

    @staticmethod
    def _normalize_logical_operators(template: str) -> str:
        """
        Normalize RFC 9535 logical operators to jsonpath2 text operators.

        RFC 9535 standard defines: &&, ||, !
        jsonpath2 library uses text operators: and, or, not
        This method normalizes symbol operators to text operators.

        Args:
            template: JSONPath template string

        Returns:
            Normalized template with text logical operators
        """
        result = []
        i = 0
        in_string = False
        string_char = None

        while i < len(template):
            char = template[i]

            # Track if we're inside a string literal
            if char in ('"', "'") and (i == 0 or template[i-1] != '\\'):
                if not in_string:
                    in_string = True
                    string_char = char
                elif char == string_char:
                    in_string = False
                    string_char = None

            if not in_string:
                # Replace && with and
                if char == '&' and i + 1 < len(template) and template[i + 1] == '&':
                    result.append(' and ')
                    i += 2
                    continue

                # Replace || with or
                elif char == '|' and i + 1 < len(template) and template[i + 1] == '|':
                    result.append(' or ')
                    i += 2
                    continue

                # Replace ! with not (but not in !=)
                elif char == '!' and i + 1 < len(template) and template[i + 1] != '=':
                    result.append('not ')
                    i += 1
                    continue

            result.append(char)
            i += 1

        return ''.join(result)

    @staticmethod
    def _add_parentheses_to_filter(template: str) -> str:
        """
        Add parentheses around filter expressions if not present.

        jsonpath2 library requires parentheses: $[?(@.age > 25)] not $[?@.age > 25]
        RFC 9535 allows both syntaxes, so we normalize to jsonpath2 format.

        Args:
            template: JSONPath template string

        Returns:
            Template with parentheses added
        """
        result = template

        # Find all [?@ patterns that don't have ( after ?
        pattern = r'\[\?(?!\()'  # [? not followed by (
        positions = []
        for match in re.finditer(pattern, result):
            # Check if next char is @ or space then @
            pos = match.end()
            if pos < len(result) and (result[pos] == '@' or (result[pos] == ' ' and pos + 1 < len(result) and result[pos + 1] == '@')):
                positions.append(match.start())

        # Process from right to left to maintain positions
        for pos in reversed(positions):
            # Find the matching ]
            depth = 1
            i = pos + 2  # After [?
            closing_pos = None

            while i < len(result) and depth > 0:
                if result[i] == '[':
                    depth += 1
                elif result[i] == ']':
                    depth -= 1
                    if depth == 0:
                        closing_pos = i
                        break
                i += 1

            if closing_pos is not None:
                # Insert ) before ]
                result = result[:closing_pos] + ')' + result[closing_pos:]
                # Insert ( after ?
                insert_pos = pos + 2  # After [?
                result = result[:insert_pos] + '(' + result[insert_pos:]

        return result

    def compile(self, jsonpath_str: str) -> str:
        """
        Compile JSONPath expression to raw SQL query.

        Args:
            jsonpath_str: JSONPath expression string

        Returns:
            Raw SQL query string

        Example:
            >>> schema = SchemaDef(...)
            >>> compiler = JSONPathToRawSQLCompiler(schema)
            >>> sql = compiler.compile("$.users[?(@.age > 18)].orders")
            >>> print(sql)
            SELECT orders.*
            FROM users
            JOIN orders ON users.user_id = orders.id
            WHERE users.age > 18
        """
        # Add parentheses to filter expressions (required by jsonpath2)
        normalized_jsonpath = self._add_parentheses_to_filter(jsonpath_str)

        # Normalize == to = for jsonpath2 library compatibility
        normalized_jsonpath = self._normalize_equality_operator(normalized_jsonpath)

        # Normalize logical operators for jsonpath2 library compatibility
        normalized_jsonpath = self._normalize_logical_operators(normalized_jsonpath)

        # Parse JSONPath
        path = Path.parse_str(normalized_jsonpath)

        # Create compilation context
        context = CompilationContext(self.schema)

        # Visit AST and compile to SQL
        visitor = get_visitor(path.root_node)
        visitor.visit(path.root_node, context)

        # Build final SQL query
        return context.query.build()



"""Deferred pattern interfaces."""
from typing import Any, Callable, Generic, Protocol, TypeVar

T = TypeVar("T")

DeferredCallback = Callable[[T], Exception | None]
"""Callback function that processes a value and may return an error."""


class IDeferred(Protocol[T]):
    """Interface for deferred operations (similar to Promise)."""

    def then(
        self,
        on_success: DeferredCallback[T],
        on_error: DeferredCallback[Exception],
    ) -> "IDeferred[Any]":
        """
        Register callbacks for success and error cases.

        Args:
            on_success: Callback to execute on successful resolution
            on_error: Callback to execute on rejection

        Returns:
            New Deferred for chaining
        """
        ...



import typing
from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Hashable

from hypothesis import strategies

from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.observable.interfaces import IObservable


__all__ = (
    'INameable',
    'IShunt',
    'ICloneable',
    'ISetupable',
    'IProvidable',
    'IInputOutput',
    'IValueProvider',
    'ICompositeInputOutput',
    'ICompositeValueProvider',
    'IEntityProvider',
    'IReferenceProvider',
    'IDependentInputOutput',
    'IDependentProvider',
    'IRelativeProvider',
    'IValueGenerator',
    'IValueAnyGenerator',
)

T_Input = typing.TypeVar("T_Input")
T_Output = typing.TypeVar("T_Output")
T_Cloneable = typing.TypeVar("T_Cloneable")
T_Id_Output = typing.TypeVar("T_Id_Output")


class INameable(metaclass=ABCMeta):

    @property
    @abstractmethod
    def provider_name(self) -> str:
        raise NotImplementedError

    @provider_name.setter
    @abstractmethod
    def provider_name(self, value: str):
        raise NotImplementedError


class IShunt(metaclass=ABCMeta):

    @abstractmethod
    def __getitem__(self, key: typing.Hashable) -> typing.Any:
        raise NotImplementedError

    @abstractmethod
    def __setitem__(self, key: typing.Hashable, value: typing.Any):
        raise NotImplementedError

    @abstractmethod
    def __contains__(self, key: typing.Hashable):
        raise NotImplementedError


class ICloneable(metaclass=ABCMeta):

    @abstractmethod
    def empty(self, shunt: IShunt | None = None) -> typing.Self:
        # For older python: def empty(self: T_Cloneable, shunt: IShunt | None = None) -> T_Cloneable:
        raise NotImplementedError

    @abstractmethod
    def do_empty(self, clone: typing.Self, shunt: IShunt):
        raise NotImplementedError


class ISetupable(metaclass=ABCMeta):

    @abstractmethod
    async def setup(self, session: ISession):
        raise NotImplementedError

    @abstractmethod
    async def cleanup(self, session: ISession):
        raise NotImplementedError


class IProvidable(metaclass=ABCMeta):

    @abstractmethod
    def reset(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def populate(self, session: ISession) -> None:
        raise NotImplementedError

    @abstractmethod
    def is_complete(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_transient(self) -> bool:
        raise NotImplementedError


class IInputOutput(typing.Generic[T_Input, T_Output], metaclass=ABCMeta):

    @abstractmethod
    async def create(self, session: ISession) -> T_Output:
        raise NotImplementedError

    @abstractmethod
    def set(self, value: T_Input) -> None:
        raise NotImplementedError

    @abstractmethod
    def get(self) -> T_Input:
        raise NotImplementedError

    @abstractmethod
    async def append(self, session: ISession, value: T_Output):
        raise NotImplementedError


class IValueProvider(
    IInputOutput[T_Input, T_Output], IProvidable, IObservable, INameable, ICloneable,
    ISetupable, typing.Generic[T_Input, T_Output], metaclass=ABCMeta
):
    """
    Immutable.
    Architecture:
    IValueProvider = f(input_value | None) = result,
    where
    result : T <- Distributor[T] <- (
        <- result : result ∈ Sᴛ ∧ P(specification) ~ 𝒟(S)  # select from a set with given probability distribution and Specification
        or
        <- result <- result_factory(input_value)
            <- input_value <- (
                set(value)
                or
                ValueGenerator(position | None) <- position | None
            )
        ),
    where
        ":" means instance of type,
        "<-" means "from",
        "∈" means belongs,
        "Sᴛ" or "{x : T}" means set of type "T",
        "∧" means satisfies the condition P(),
        "~ 𝒟(S)" means according to the probability distribution,
        "Σx" means composition of "x",
        "⊆" means subset of a composition.
    """
    pass


class ICompositeInputOutput(typing.Generic[T_Input, T_Output], metaclass=ABCMeta):
    """
    Структура Provider не совпадает со структурой агрегата, если агрегат приводится в требуемое состояние многоходово
    (см. агрегат Specialist at grade project).
    Это подсказка на вопрос о том, должен ли Distributor хранить сырые значения провайдера или готовый агрегат.

    В method self.set(...) технически невозможно установить в качестве значения итоговый тип,
    т.к. для валидного его состояния банально может не хватать данных (Auto Increment PK, FK).
    """

    @abstractmethod
    async def create(self, session: ISession) -> T_Output:
        raise NotImplementedError

    @abstractmethod
    def set(self, value: T_Input) -> None:
        """
        Не используем **kwargs, т.к. иначе придется инспектировать сигнатуру каждого вложенного сеттера
        (композиция может быть вложенной).
        Ну и в принципе здесь можно принимать Specification вторым аргументом.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self) -> T_Input:
        raise NotImplementedError

    @abstractmethod
    async def append(self, session: ISession, value: T_Output):
        raise NotImplementedError


class ICompositeValueProvider(
    IInputOutput[T_Input, T_Output], IProvidable, IObservable, INameable, ICloneable,
    ISetupable, typing.Generic[T_Input, T_Output], metaclass=ABCMeta
):
    """
    Immutable. Composite ValueObject.
    Architecture:
    ICompositeValueProvider = f(Σ input_value | None) = result,
    where
    result : T <- Distributor[T] <- (
        <- result : result ∈ Sᴛ ∧ P(specification) ~ 𝒟(S)  # select from a set with given probability distribution and Specification
        or
        <- result <- result_factory(Σ leaf_result)
            <- Σ IValueProvider(∈ Σ input_value) | ICompositeValueProvider(⊆ Σ input_value)
    ),
    where
        ":" means instance of type,
        "<-" means "from",
        "∈" means belongs,
        "Sᴛ" or "{x : T}" means set of type "T",
        "∧" means satisfies the condition P(),
        "~ 𝒟(S)" means according to the probability distribution,
        "Σx" means composition of "x",
        "⊆" means subset of a composition.
    """
    pass


class IEntityProvider(
    ICompositeInputOutput[T_Input, T_Output], IProvidable, IObservable, INameable, ICloneable,
    ISetupable, typing.Generic[T_Input, T_Output], metaclass=ABCMeta
):
    """
    Mutable. Saved as part of aggregate.
    """

    @abstractmethod
    def on_init(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def id_provider(self) -> IValueProvider[T_Input, T_Output]:
        raise NotImplementedError


class IReferenceProvider(
    IValueProvider[T_Input, T_Id_Output],
    typing.Generic[T_Input, T_Output, T_Id_Output], metaclass=ABCMeta
):

    @property
    @abstractmethod
    def aggregate_provider(self) -> IEntityProvider[T_Input, T_Output]:
        raise NotImplementedError

    @aggregate_provider.setter
    @abstractmethod
    def aggregate_provider(
            self,
            aggregate_provider: IEntityProvider[T_Input, T_Output] | Callable[[], IEntityProvider[T_Input, T_Output]]
    ) -> None:
        raise NotImplementedError


class IDependentInputOutput(typing.Generic[T_Input, T_Output], metaclass=ABCMeta):

    @abstractmethod
    async def create(self, session: ISession) -> list[T_Output]:
        raise NotImplementedError

    @abstractmethod
    def set(self, value: list[T_Input], weights: list[float] | None = None) -> None:
        raise NotImplementedError

    @abstractmethod
    def get(self) -> list[T_Input]:
        raise NotImplementedError


class IDependentProvider(
    IDependentInputOutput[T_Input, T_Id_Output], IProvidable, IObservable, INameable, ICloneable,
    ISetupable, typing.Generic[T_Input, T_Output, T_Id_Output], metaclass=ABCMeta
):
    """
    Я думал над тем, чтоб разбить providers на m2o и o2m, но это было бы неуместно потому,
    что, например, для генерации значения зарплаты мы можем использовать IO2MDistributor,
    но это не o2m, это, по сути, m2o.

    Вместо m2o и o2m можно было бы использовать термины belongs и has,
    но они неуместны по отношению к простым значениями. Не может User принадлежать status.
    """

    @property
    @abstractmethod
    def aggregate_providers(self) -> list[IEntityProvider[T_Input, T_Output]]:
        raise NotImplementedError

    @aggregate_providers.setter
    @abstractmethod
    def aggregate_providers(
            self,
            aggregate_provider: list[IEntityProvider[T_Input, T_Output] |
                                     Callable[[], IEntityProvider[T_Input, T_Output]]]
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_dependency_id(self, dependency_id: typing.Any) -> None:
        raise NotImplementedError


class IRelativeProvider(metaclass=ABCMeta):

    @abstractmethod
    def set_scope(self, scope: Hashable) -> None:
        raise NotImplementedError


class IValueGenerator(typing.Generic[T_Input], metaclass=ABCMeta):
    """
    Фабрика значений для дистрибьюторов.
    Принимает session и опциональный position (номер в последовательности).
    """

    @abstractmethod
    async def __call__(self, session: ISession, position: int | None = None) -> T_Input:
        raise NotImplementedError


IValueAnyGenerator: typing.TypeAlias = (
    IValueGenerator[T_Input] | typing.Iterable[T_Input] | strategies.SearchStrategy[T_Input] | Callable
)



import typing

from ascetic_ddd.faker.domain.providers._mixins import BaseProvider
from ascetic_ddd.faker.domain.providers.interfaces import IValueProvider
from ascetic_ddd.faker.domain.values.empty import empty, Empty
from ascetic_ddd.faker.domain.session.interfaces import ISession

__all__ = ('PredefinedProvider',)

T_Input = typing.TypeVar("T_Input")
T_Output = typing.TypeVar("T_Output")


class PredefinedProvider(BaseProvider[T_Input, T_Output], IValueProvider[T_Input, T_Output]):
    """
    TODO: Remove me. Use ValueProvaider(
        distributor= DummyDistributor(),
        value_generator=lambda session, position: 'predefined_value',
        result_factory=result_factory,
        result_exporter=result_exporter,
    ) instead.
    """
    _default_input_value: T_Input | Empty = empty
    _provider_name: str | None = None

    def __init__(
            self,
            default_value: T_Input | Empty = empty,
            result_factory: typing.Callable[[T_Input], T_Output] | None = None,
    ):
        super().__init__()

        self._default_input_value = default_value

        if result_factory is None:
            def result_factory(result):
                return result

        self._result_factory = result_factory

    async def create(self, session: ISession) -> T_Output:
        return self._result_factory(self._input_value)

    async def populate(self, session: ISession) -> None:
        if self._input_value is empty:
            self.set(self._default_input_value)



import typing

from ascetic_ddd.faker.domain.distributors.m2o import DummyDistributor
from ascetic_ddd.faker.domain.distributors.m2o.interfaces import ICursor, IM2ODistributor
from ascetic_ddd.faker.domain.providers._mixins import BaseCompositeDistributionProvider
from ascetic_ddd.faker.domain.specification.empty_specification import EmptySpecification
from ascetic_ddd.faker.domain.values.empty import empty
from ascetic_ddd.faker.domain.specification.object_pattern_specification import ObjectPatternSpecification
from ascetic_ddd.faker.domain.session.interfaces import ISession

__all__ = (
    'CompositeValueProvider',
)


T_Input = typing.TypeVar("T_Input")
T_Output = typing.TypeVar("T_Output")


class CompositeValueProvider(
    BaseCompositeDistributionProvider,
    typing.Generic[T_Input, T_Output]
):
    _result_factory: typing.Callable[[...], T_Output] = None  # T_Output of each nested Provider.
    _result_exporter: typing.Callable[[T_Output], T_Input] = None

    def __init__(
            self,
            distributor: IM2ODistributor[T_Input] | None = None,
            result_factory: typing.Callable[[...], T_Output] | None = None,  # T_Output of each nested Provider.
            result_exporter: typing.Callable[[T_Output], T_Input] | None = None,
    ):
        if distributor is None:
            distributor = DummyDistributor()

        if self._result_factory is None:
            if result_factory is None:

                def result_factory(**kwargs):
                    return kwargs

            self._result_factory = result_factory

        if self._result_exporter is None:
            if result_exporter is None:

                def result_exporter(value):
                    return value

            self._result_exporter = result_exporter

        super().__init__(distributor=distributor)
        self.on_init()

    def on_init(self):
        pass

    async def create(self, session: ISession) -> T_Output:
        return self._output_result

    async def populate(self, session: ISession) -> None:
        if self.is_complete():
            if self._output_result is empty:
                self._output_result = await self._default_factory(session)
            return

        if self._input_value is empty:
            specification = EmptySpecification()
        else:
            specification = ObjectPatternSpecification(self._input_value, self._result_exporter)

        await self.do_populate(session)
        cursors = {}
        for attr, provider in self._providers.items():
            try:
                await provider.populate(session)
            except ICursor as cursor:
                cursors[attr] = cursor

        try:
            result = await self._distributor.next(session, specification)
            if result is not None:
                value = self._result_exporter(result)
                self.set(value)
            else:
                self.set(None)
            # self.set() could reset self._output_result
            self._output_result = result
        except ICursor as cursor:
            result = await self._default_factory(session, cursor.position)
            value = self._result_exporter(result)
            self.set(value)
            # self.set() could reset self._output_result
            self._output_result = result
            if not self.is_transient():
                await cursor.append(session, self._output_result)
            # infinite recursion
            # await self.populate(session)

    async def _default_factory(self, session: ISession, position: typing.Optional[int] = None):
        data = dict()
        for attr, provider in self._providers.items():
            data[attr] = await provider.create(session)
        return self._result_factory(**data)

    async def do_populate(self, session: ISession) -> None:
        pass



"""AST nodes for Specification pattern."""
from abc import ABC, abstractmethod
from typing import Any, Callable, Protocol

from .constants import OPERATOR, ASSOCIATIVITY


class Visitor(Protocol):
    """Visitor interface for traversing AST nodes."""

    def visit_global_scope(self, node: "GlobalScope") -> None:
        """Visit a global scope node."""
        ...

    def visit_object(self, node: "Object") -> None:
        """Visit an object node."""
        ...

    def visit_collection(self, node: "Collection") -> None:
        """Visit a collection node."""
        ...

    def visit_item(self, node: "Item") -> None:
        """Visit an item node."""
        ...

    def visit_field(self, node: "Field") -> None:
        """Visit a field node."""
        ...

    def visit_value(self, node: "Value") -> None:
        """Visit a value node."""
        ...

    def visit_prefix(self, node: "Prefix") -> None:
        """Visit a prefix node."""
        ...

    def visit_infix(self, node: "Infix") -> None:
        """Visit an infix node."""
        ...

    def visit_postfix(self, node: "Postfix") -> None:
        """Visit a prefix node."""
        ...


class Visitable(ABC):
    """Base class for visitable nodes."""

    @abstractmethod
    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        pass


class Operable(Protocol):
    """Interface for operable nodes."""

    def associativity(self) -> ASSOCIATIVITY:
        """Return operator associativity."""
        ...

    def operator(self) -> OPERATOR:
        """Return operator type."""
        ...


class EmptiableObject(Protocol):
    """Interface for objects that can be empty (scopes)."""

    def parent(self) -> "EmptiableObject":
        """Return parent object."""
        ...

    def name(self) -> str:
        """Return object name."""
        ...

    def is_root(self) -> bool:
        """Check if this is a root object."""
        ...

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        ...


class Value(Visitable):
    """Node representing a constant value."""

    def __init__(self, value: Any):
        self._value = value

    def value(self) -> Any:
        """Return the stored value."""
        return self._value

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_value(self)


class Prefix(Visitable):
    """Node representing a prefix operator (e.g., NOT)."""

    def __init__(
        self,
        operator: OPERATOR,
        operand: Visitable,
        associativity: ASSOCIATIVITY,
    ):
        self._operator = operator
        self._operand = operand
        self._associativity = associativity

    def operand(self) -> Visitable:
        """Return the operand."""
        return self._operand

    def operator(self) -> OPERATOR:
        """Return the operator."""
        return self._operator

    def associativity(self) -> ASSOCIATIVITY:
        """Return operator associativity."""
        return self._associativity

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_prefix(self)


class Not(Prefix):
    def __init__(self, operand: Visitable):
        super().__init__(OPERATOR.NOT, operand, ASSOCIATIVITY.RIGHT_ASSOCIATIVE)


class Infix(Visitable):
    """Node representing an infix operator (e.g., AND, OR, =, >)."""

    def __init__(
        self,
        left: Visitable,
        operator: OPERATOR,
        right: Visitable,
        associativity: ASSOCIATIVITY,
    ):
        self._left = left
        self._operator = operator
        self._right = right
        self._associativity = associativity

    def left(self) -> Visitable:
        """Return the left operand."""
        return self._left

    def operator(self) -> OPERATOR:
        """Return the operator."""
        return self._operator

    def right(self) -> Visitable:
        """Return the right operand."""
        return self._right

    def associativity(self) -> ASSOCIATIVITY:
        """Return operator associativity."""
        return self._associativity

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_infix(self)


class Equal(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.EQ, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class NotEqual(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.NE, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class Is(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.IS, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class GreaterThan(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.GT, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class LessThan(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.LT, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class GreaterThanEqual(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.GTE, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class LessThanEqual(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.LTE, right, ASSOCIATIVITY.NON_ASSOCIATIVE)


class And(Infix):
    def __init__(self, left: Visitable, *rights: Visitable):
        if not rights:
            raise ValueError("At least one right operand is required")
        left_folded, right_folded = _fold_rights(type(self), left, *rights)
        super().__init__(left_folded, OPERATOR.AND, right_folded, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class Or(Infix):
    def __init__(self, left: Visitable, *rights: Visitable):
        if not rights:
            raise ValueError("At least one right operand is required")
        left_folded, right_folded = _fold_rights(type(self), left, *rights)
        super().__init__(left_folded, OPERATOR.OR, right_folded, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class LeftShift(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.LSHIFT, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class RightShift(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.RSHIFT, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class Add(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.ADD, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class Sub(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.SUB, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class Mul(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.MUL, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class Div(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.DIV, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


class Mod(Infix):
    def __init__(self, left: Visitable, right: Visitable):
        super().__init__(left, OPERATOR.MOD, right, ASSOCIATIVITY.LEFT_ASSOCIATIVE)


def _fold_rights(
    callable_func: Callable[..., Infix],
    left: Visitable,
    *rights: Visitable,
) -> tuple[Visitable, Visitable]:
    """Fold multiple right operands into nested binary operations."""
    rights_list = list(rights)
    while len(rights_list) > 1:
        left = callable_func(left, rights_list.pop(0))
    return left, rights_list[0]


class Postfix(Visitable):
    """Node representing a prefix operator (e.g., NOT)."""

    def __init__(
        self,
        operand: Visitable,
        operator: OPERATOR,
        associativity: ASSOCIATIVITY,
    ):
        self._operand = operand
        self._operator = operator
        self._associativity = associativity

    def operand(self) -> Visitable:
        """Return the operand."""
        return self._operand

    def operator(self) -> OPERATOR:
        """Return the operator."""
        return self._operator

    def associativity(self) -> ASSOCIATIVITY:
        """Return operator associativity."""
        return self._associativity

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_postfix(self)


class IsNull(Postfix):
    def __init__(self, operand: Visitable):
        super().__init__(operand, OPERATOR.IS_NULL, ASSOCIATIVITY.NON_ASSOCIATIVE)


class IsNotNull(Postfix):
    def __init__(self, operand: Visitable):
        super().__init__(operand, OPERATOR.IS_NOT_NULL, ASSOCIATIVITY.NON_ASSOCIATIVE)


class GlobalScope(Visitable):
    """Node representing the global scope (root)."""

    def parent(self) -> "GlobalScope":
        """Return parent (self for root)."""
        return self

    def name(self) -> str:
        """Return the name."""
        return "Empty"

    def is_root(self) -> bool:
        """Check if this is root."""
        return True

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_global_scope(self)


class Object(Visitable):
    """Node representing an object in the scope chain."""

    def __init__(self, parent: EmptiableObject, name: str):
        self._parent = parent
        self._name = name

    def parent(self) -> EmptiableObject:
        """Return the parent object."""
        return self._parent

    def name(self) -> str:
        """Return the object name."""
        return self._name

    def is_root(self) -> bool:
        """Check if this is root."""
        return False

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_object(self)


class Collection(Visitable):
    """Node representing a collection with a wildcard and predicate."""

    def __init__(self, parent: EmptiableObject, name: str, predicate: Visitable):
        self._parent = parent
        self._name = name
        self._predicate = predicate

    def parent(self) -> EmptiableObject:
        """Return the parent object."""
        return self._parent

    def name(self) -> str:
        """Return the collection name (*)."""
        return self._name

    def is_root(self) -> bool:
        """Check if this is root."""
        return False

    def predicate(self) -> Visitable:
        """Return the predicate for filtering collection items."""
        return self._predicate

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_collection(self)


class Wildcard(Collection):
    def __init__(self, parent: EmptiableObject, predicate: Visitable):
        super().__init__(parent, "*", predicate)


class Item(Visitable):
    """Node representing the current item in a collection (@)."""

    def parent(self) -> EmptiableObject:
        """Return parent (global scope)."""
        return GlobalScope()

    def name(self) -> str:
        """Return the item name (@)."""
        return "@"

    def is_root(self) -> bool:
        """Check if this is root."""
        return True

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_item(self)


class Field(Visitable):
    """Node representing a field access."""

    def __init__(self, obj: EmptiableObject, name: str):
        self._object = obj
        self._name = name

    def name(self) -> str:
        """Return the field name."""
        return self._name

    def object(self) -> EmptiableObject:
        """Return the object containing this field."""
        return self._object

    def accept(self, visitor: Visitor) -> None:
        """Accept a visitor."""
        visitor.visit_field(self)


def extract_field_path(node: Field) -> list[str]:
    """Extract the full path to a field as a list of names."""
    path = [node.name()]
    obj: EmptiableObject = node.object()
    while not obj.is_root():
        path.insert(0, obj.name())
        obj = obj.parent()
    return path



import base64

try:
    import cPickle as pickle
except ImportError:
    import pickle

__all__ = ('serialize', 'deserialize')


def serialize(val):
    return base64.standard_b64encode(
        pickle.dumps(val, protocol=pickle.HIGHEST_PROTOCOL)
    ).decode('utf-8')


def deserialize(val):
    return pickle.loads(base64.standard_b64decode(val.encode('utf-8')))



"""See https://developer.mozilla.org/en-US/docs/Web/API/GeolocationCoordinates."""
from abc import ABCMeta, abstractmethod
from decimal import Decimal

try:
    from geopy.distance import geodesic
except ImportError:
    geodesic = None

from .point import Point

__all__ = (
    "GeolocationCoordinates",
    "IGeolocationCoordinatesExporter",
)


class GeolocationCoordinates:
    _latitude: Decimal
    _longitude: Decimal
    _altitude: Decimal | None = None
    _accuracy: Decimal | None = None
    _altitude_accuracy: Decimal | None = None
    _heading: Decimal | None = None
    _speed: Decimal | None = None

    def __init__(
        self,
        latitude: Decimal,
        longitude: Decimal,
        altitude: Decimal | None = None,
        accuracy: Decimal | None = None,
        altitude_accuracy: Decimal | None = None,
        heading: Decimal | None = None,
        speed: Decimal | None = None,
    ) -> None:
        self._latitude = latitude
        self._longitude = longitude
        self._altitude = altitude
        self._accuracy = accuracy
        self._altitude_accuracy = altitude_accuracy
        self._heading = heading
        self._speed = speed

    @property
    def point(self) -> Point:
        # Geopy cannot calculate distance with altitude
        return Point(self._latitude, self._longitude, None)

    def distance(self, location: Point):
        return geodesic(self.point, location).m - (float(self._accuracy) if self._accuracy else 0)

    def export(self, exporter: "IGeolocationCoordinatesExporter") -> None:
        exporter.set_latitude(self._latitude)
        exporter.set_longitude(self._longitude)
        exporter.set_altitude(self._altitude)
        exporter.set_accuracy(self._accuracy)
        exporter.set_altitude_accuracy(self._altitude_accuracy)
        exporter.set_heading(self._heading)
        exporter.set_speed(self._speed)


class IGeolocationCoordinatesExporter(metaclass=ABCMeta):

    @abstractmethod
    def set_longitude(self, value: Decimal) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_latitude(self, value: Decimal) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_altitude(self, value: Decimal | None) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_accuracy(self, value: Decimal | None) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_altitude_accuracy(self, value: Decimal | None) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_heading(self, value: Decimal | None) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_speed(self, value: Decimal | None) -> None:
        raise NotImplementedError



from ascetic_ddd.faker.infrastructure.distributors.m2o.pg_weighted_distributor import PgWeightedDistributor
from ascetic_ddd.faker.infrastructure.distributors.m2o.pg_skew_distributor import PgSkewDistributor
from ascetic_ddd.faker.infrastructure.distributors.m2o.pg_sequence_distributor import PgSequenceDistributor
from ascetic_ddd.faker.infrastructure.distributors.m2o.factory import pg_distributor_factory



import copy
import functools
import typing
import abc
import logging
from collections.abc import Hashable, Callable

from ascetic_ddd.disposable import IDisposable
from ascetic_ddd.faker.domain.distributors.m2o.interfaces import IM2ODistributor
from ascetic_ddd.faker.domain.providers.interfaces import (
    IValueProvider, INameable, IShunt, ICloneable,
    ICompositeValueProvider, IValueGenerator, IDependentProvider
)
from ascetic_ddd.faker.domain.providers.value_generators import prepare_value_generator
from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.faker.domain.specification.interfaces import ISpecification
from ascetic_ddd.faker.domain.values.empty import empty, Empty
from ascetic_ddd.observable.interfaces import IObservable
from ascetic_ddd.observable.observable import Observable

__all__ = (
    'ObservableMixin',
    'NameableMixin',
    'CloneableMixin',
    'Shunt',
    'BaseProvider',
    'BaseDistributionProvider',
    'BaseCompositeProvider',
    'BaseCompositeDistributionProvider',
)

T_Input = typing.TypeVar("T_Input")
T_Output = typing.TypeVar("T_Output")
T_Cloneable = typing.TypeVar("T_Cloneable")


class ObservableMixin(Observable, IObservable, metaclass=abc.ABCMeta):

    _aspect_mapping = {
        "distributor": "_distributor"
    }

    def _split_aspect(self, aspect: typing.Hashable) -> tuple[str | None, typing.Hashable]:
        if isinstance(aspect, str) and "." in aspect:
            attr, inner_aspect = aspect.split('.', maxsplit=1)
            attr = self._aspect_mapping.get(attr, attr)
            return attr, inner_aspect
        return None, aspect

    def attach(self, aspect: Hashable, observer: Callable, id_: Hashable | None = None) -> IDisposable:
        attr, inner_aspect = self._split_aspect(aspect)
        if attr is not None:
            return getattr(self, attr).attach(inner_aspect, observer, id_)
        else:
            return super().attach(inner_aspect, observer, id_)

    def detach(self, aspect: Hashable, observer: Callable, id_: Hashable | None = None):
        attr, inner_aspect = self._split_aspect(aspect)
        if attr is not None:
            return getattr(self, attr).detach(inner_aspect, observer, id_)
        else:
            super().detach(inner_aspect, observer, id_)

    def notify(self, aspect: Hashable, *args, **kwargs):
        attr, inner_aspect = self._split_aspect(aspect)
        if attr is not None:
            return getattr(self, attr).notify(inner_aspect, *args, **kwargs)
        else:
            super().notify(inner_aspect, *args, **kwargs)

    async def anotify(self, aspect: Hashable, *args, **kwargs):
        attr, inner_aspect = self._split_aspect(aspect)
        if attr is not None:
            return await getattr(self, attr).anotify(inner_aspect, *args, **kwargs)
        else:
            await super().anotify(inner_aspect, *args, **kwargs)


class NameableMixin(INameable, metaclass=abc.ABCMeta):
    _provider_name: str | None = None

    @property
    def provider_name(self) -> str:
        return self._provider_name

    @provider_name.setter
    def provider_name(self, value: str):
        if self._provider_name is None:
            self._provider_name = value


class Shunt(IShunt):

    def __init__(self):
        self._data = {}

    def __getitem__(self, key: typing.Hashable) -> typing.Any:
        return self._data[key]

    def __setitem__(self, key: typing.Hashable, value: typing.Any):
        self._data[key] = value

    def __contains__(self, key: typing.Hashable):
        return key in self._data


class CloneableMixin(ICloneable):

    def empty(self, shunt: IShunt | None = None) -> typing.Self:
        if shunt is None:
            shunt = Shunt()
        if self in shunt:
            return shunt[self]
        c = copy.copy(self)
        self.do_empty(c, shunt)
        shunt[self] = c
        return c

    def do_empty(self, clone: typing.Self, shunt: IShunt):
        pass


class BaseProvider(
    NameableMixin,
    ObservableMixin,
    CloneableMixin,
    IValueProvider[T_Input, T_Output],
    typing.Generic[T_Input, T_Output],
    metaclass=abc.ABCMeta
):
    _input_value: T_Input | Empty = empty
    _output_result: T_Output | Empty = empty

    def reset(self) -> None:
        self._input_value = empty
        self._output_result = empty
        self.notify('input_value', self._input_value)

    def set(self, value: T_Input) -> None:
        if self._input_value != value:
            self._input_value = value
            self._output_result = empty
            self.notify('input_value', self._input_value)

    def get(self) -> T_Input:
        return self._input_value

    def do_empty(self, clone: typing.Self, shunt: IShunt):
        clone._input_value = empty
        clone._output_result = empty

    def is_complete(self) -> bool:
        return self._output_result is not empty

    def is_transient(self) -> bool:
        return self._input_value is empty

    async def append(self, session: ISession, value: T_Output):
        pass

    async def setup(self, session: ISession):
        pass

    async def cleanup(self, session: ISession):
        pass


class BaseDistributionProvider(BaseProvider[T_Input, T_Output], typing.Generic[T_Input, T_Output],
                               metaclass=abc.ABCMeta):
    _distributor: IM2ODistributor[T_Output]

    def __init__(self, distributor: IM2ODistributor):
        self._distributor = distributor
        super().__init__()

    @property
    def provider_name(self) -> str:
        return self._provider_name

    @provider_name.setter
    def provider_name(self, value: str):
        self._provider_name = value
        self._distributor.provider_name = value

    async def setup(self, session: ISession):
        await self._distributor.setup(session)
        await super().setup(session)

    async def cleanup(self, session: ISession):
        await self._distributor.cleanup(session)
        await super().cleanup(session)

    async def append(self, session: ISession, value: T_Input):
        await self._distributor.append(session, value)


class BaseCompositeProvider(
    ObservableMixin,
    CloneableMixin,
    ICompositeValueProvider[T_Input, T_Output],
    typing.Generic[T_Input, T_Output],
    metaclass=abc.ABCMeta
):

    _input_value: T_Input | Empty = empty
    _output_result: T_Output | Empty = empty
    _provider_name: str | None = None

    def is_complete(self) -> bool:
        return (
            self._output_result is not empty or
            all(provider.is_complete() for provider in self._providers.values())
        )

    def is_transient(self) -> bool:
        return any(provider.is_transient() for provider in self._providers.values())

    def do_empty(self, clone: typing.Self, shunt: IShunt):
        clone._input_value = empty
        clone._output_result = empty
        for attr, provider in self._providers.items():
            setattr(clone, attr, provider.empty(shunt))
        clone.on_init()

    def reset(self) -> None:
        self._input_value = empty
        self._output_result = empty
        self.notify('input_value', self._input_value)
        for provider in self._providers.values():
            provider.reset()

    def set(self, value: T_Input) -> None:
        """
        https://docs.python.org/3/library/stdtypes.html#mapping-types-dict
        """
        if self._input_value != value:
            self._input_value = value
            self._output_result = empty
            self.notify('input_value', value)
        if value is not empty:
            for attr, val in value.items():
                """
                Вложенная композиция поддерживается автоматически.
                """
                getattr(self, attr).set(val)

    def get(self) -> T_Input:
        value = dict()
        for attr, provider in self._providers.items():
            val = provider.get()
            if val is not empty:
                value[attr] = val
        return value

    async def append(self, session: ISession, value: T_Output):
        pass

    async def setup(self, session: ISession):
        for provider in self._providers.values():
            await provider.setup(session)
        for provider in self._dependent_providers.values():
            await provider.setup(session)

    async def cleanup(self, session: ISession):
        for provider in self._providers.values():
            await provider.cleanup(session)
        for provider in self._dependent_providers.values():
            await provider.cleanup(session)

    @classmethod
    @property
    @functools.cache
    def _provider_attrs(cls) -> list[str]:
        attrs = list()
        for cls_ in cls.mro():  # Use self.__dict__ or self.__reduce__() instead?
            if hasattr(cls_, '__annotations__'):
                for key, type_hint in cls_.__annotations__.items():
                    if not key.startswith('_') and key not in attrs:
                        # Skip IDependentProvider - it's handled separately
                        origin = typing.get_origin(type_hint) or type_hint
                        if isinstance(origin, type) and issubclass(origin, IDependentProvider):
                            continue
                        attrs.append(key)
        return attrs

    @classmethod
    @property
    @functools.cache
    def _dependent_provider_attrs(cls) -> list[str]:
        """Returns attribute names that are IDependentProvider."""
        attrs = list()
        for cls_ in cls.mro():
            if hasattr(cls_, '__annotations__'):
                for key, type_hint in cls_.__annotations__.items():
                    if not key.startswith('_') and key not in attrs:
                        origin = typing.get_origin(type_hint) or type_hint
                        if isinstance(origin, type) and issubclass(origin, IDependentProvider):
                            attrs.append(key)
        return attrs

    @property
    def _providers(self) -> dict[str, IValueProvider[typing.Any, typing.Any]]:
        return {i: getattr(self, i) for i in self._provider_attrs}

    @property
    def _dependent_providers(self) -> dict[str, IDependentProvider[typing.Any, typing.Any, typing.Any]]:
        return {i: getattr(self, i) for i in self._dependent_provider_attrs}

    @property
    def provider_name(self):
        return self._provider_name

    @provider_name.setter
    def provider_name(self, value):
        self._provider_name = value
        for attr, provider in self._providers.items():
            provider.provider_name = "%s.%s" % (value, attr)
        for attr, provider in self._dependent_providers.items():
            provider.provider_name = "%s.%s" % (value, attr)


class BaseCompositeDistributionProvider(
    BaseCompositeProvider[T_Input, T_Output],
    typing.Generic[T_Input, T_Output],
    metaclass=abc.ABCMeta
):

    _input_value: T_Input | Empty = empty
    _output_result: T_Output | Empty = empty
    _provider_name: str | None = None
    _distributor: IM2ODistributor[T_Input]

    def __init__(self, distributor: IM2ODistributor[T_Input]):
        self._distributor = distributor
        super().__init__()

    async def append(self, session: ISession, value: T_Output):
        await self._distributor.append(session, value)
        await super().append(session, value)

    async def setup(self, session: ISession):
        await self._distributor.setup(session)
        await super().setup(session)

    async def cleanup(self, session: ISession):
        await self._distributor.cleanup(session)
        await super().cleanup(session)

    @property
    def provider_name(self):
        return self._provider_name

    @provider_name.setter
    def provider_name(self, value):
        self._provider_name = value
        self._distributor.provider_name = value
        for attr, provider in self._providers.items():
            provider.provider_name = "%s.%s" % (value, attr)



import typing
from datetime import datetime

from .adapters import Factory, Logical, Comparison, Mathematical, Nullable

__all__ = (
    'Boolean',
    'NullBoolean',
    'Number',
    'NullNumber',
    'Datetime',
    'NullDatetime',
    'Text',
    'NullText',
)

T = typing.TypeVar("T")


class Boolean(Logical, Factory[bool]):
    pass


class NullBoolean(Boolean, Nullable):
    pass


class Number(Comparison, Mathematical[T], Factory[T], typing.Generic[T]):
    pass


class NullNumber(Number[T], Nullable, typing.Generic[T]):
    pass


class Datetime(Comparison, Mathematical[datetime], Factory[datetime]):
    pass


class NullDatetime(Datetime, Nullable):
    pass


class Text(Comparison, Factory[T], typing.Generic[T]):
    pass


class NullText(Text[T], Nullable, typing.Generic[T]):
    pass



import typing

from ascetic_ddd.faker.domain.distributors.m2o.cursor import Cursor
from ascetic_ddd.observable.observable import Observable
from ascetic_ddd.faker.domain.distributors.m2o.interfaces import IM2ODistributor
from ascetic_ddd.faker.domain.session.interfaces import ISession
from ascetic_ddd.faker.domain.specification.interfaces import ISpecification

__all__ = ('DummyDistributor',)

T = typing.TypeVar("T", covariant=True)


class DummyDistributor(Observable, IM2ODistributor[T], typing.Generic[T]):
    _provider_name: str | None = None

    async def next(
            self,
            session: ISession,
            specification: ISpecification[T] | None = None,
    ) -> T:
        raise Cursor(
            position=None,
            callback=self._append,
        )

    async def _append(self, session: ISession, value: T, position: int | None):
        await self.anotify('value', session, value)

    async def append(self, session: ISession, value: T):
        await self._append(session, value, None)

    @property
    def provider_name(self):
        return self._provider_name

    @provider_name.setter
    def provider_name(self, value):
        if self._provider_name is None:
            self._provider_name = value

    async def setup(self, session: ISession):
        pass

    async def cleanup(self, session: ISession):
        pass

    def __copy__(self):
        return self

    def __deepcopy__(self, memodict={}):
        return self

    def bind_external_source(self, external_source: typing.Any) -> None:
        """DummyDistributor не использует external_source."""
        pass



"""
Native JSONPath parser for Specification Pattern without external dependencies.

Parses RFC 9535 compliant JSONPath expressions with C-style placeholders
(%s, %d, %f, %(name)s) and converts them directly to Specification AST nodes.

RFC 9535 Compliance:
- Uses == for equality (double equals)
- Uses && for logical AND (double ampersand)
- Uses || for logical OR (double pipe)
- Uses ! for logical NOT (exclamation mark)
"""
from typing import Any, Dict, Tuple, Union
import re

from ..nodes import (
    And,
    Equal,
    Field,
    GlobalScope,
    GreaterThan,
    GreaterThanEqual,
    Item,
    LessThan,
    LessThanEqual,
    Not,
    NotEqual,
    Object,
    Or,
    Value,
    Visitable,
    Wildcard,
)
from ..evaluate_visitor import Context, EvaluateVisitor


class Token:
    """Represents a token in the JSONPath expression."""

    def __init__(self, type_: str, value: Any, position: int = 0):
        self.type = type_
        self.value = value
        self.position = position

    def __repr__(self):
        return f"Token({self.type}, {self.value!r})"


class Lexer:
    """Tokenizes JSONPath expressions."""

    TOKEN_PATTERNS = [
        ("LBRACKET", r"\["),
        ("RBRACKET", r"\]"),
        ("LPAREN", r"\("),
        ("RPAREN", r"\)"),
        ("DOT", r"\."),
        ("DOLLAR", r"\$"),
        ("AT", r"@"),
        ("QUESTION", r"\?"),
        ("WILDCARD", r"\*"),
        ("AND", r"&&"),  # RFC 9535: double ampersand
        ("OR", r"\|\|"),  # RFC 9535: double pipe
        ("EQ", r"=="),  # RFC 9535: double equals (must be before single =)
        ("NE", r"!="),  # Must be before NOT to match != as one token
        ("GTE", r">="),
        ("LTE", r"<="),
        ("GT", r">"),
        ("LT", r"<"),
        ("NOT", r"!"),  # RFC 9535: exclamation mark (after !=)
        ("NUMBER", r"-?\d+\.?\d*"),
        ("STRING", r"'[^']*'|\"[^\"]*\""),
        ("PLACEHOLDER", r"%\(\w+\)[sdf]|%[sdf]"),
        ("IDENTIFIER", r"[a-zA-Z_][a-zA-Z0-9_]*"),
        ("WHITESPACE", r"\s+"),
    ]

    def __init__(self, text: str):
        self.text = text
        self.position = 0
        self.tokens = []

    def tokenize(self) -> list[Token]:
        """Tokenize the input text."""
        while self.position < len(self.text):
            matched = False

            for token_type, pattern in self.TOKEN_PATTERNS:
                regex = re.compile(pattern)
                match = regex.match(self.text, self.position)

                if match:
                    value = match.group(0)
                    if token_type != "WHITESPACE":  # Skip whitespace
                        self.tokens.append(Token(token_type, value, self.position))
                    self.position = match.end()
                    matched = True
                    break

            if not matched:
                raise SyntaxError(
                    f"Unexpected character at position {self.position}: "
                    f"{self.text[self.position]}"
                )

        return self.tokens


class NativeParametrizedSpecification:
    """
    Native JSONPath specification parser without external dependencies.

    Parses template once, binds different values at execution time.
    """

    def __init__(self, template: str):
        """
        Parse JSONPath template with placeholders.

        Args:
            template: JSONPath with %s, %d, %f or %(name)s placeholders
        """
        self.template = template
        self._placeholder_info = []
        self._placeholder_refs = []
        self._placeholder_bind_index = 0
        self._is_wildcard_context = False  # Track if we're in wildcard context

        # Extract placeholders before tokenization
        self._extract_placeholders()

    def _extract_placeholders(self):
        """Extract placeholder information from template."""
        # Find named placeholders: %(name)s, %(age)d, %(price)f
        named_pattern = r"%\((\w+)\)([sdf])"
        for match in re.finditer(named_pattern, self.template):
            name = match.group(1)
            format_type = match.group(2)
            self._placeholder_info.append(
                {
                    "name": name,
                    "format_type": format_type,
                    "positional": False,
                }
            )

        # Find positional placeholders: %s, %d, %f
        # Create a temp string without named placeholders
        temp = re.sub(named_pattern, "", self.template)
        positional_pattern = r"%([sdf])"
        position = 0
        for match in re.finditer(positional_pattern, temp):
            format_type = match.group(1)
            self._placeholder_info.append(
                {
                    "name": str(position),
                    "format_type": format_type,
                    "positional": True,
                }
            )
            position += 1

    def _parse_expression(
        self, tokens: list[Token], start: int = 0
    ) -> tuple[Visitable, int]:
        """
        Parse a filter expression from tokens.

        Args:
            tokens: List of tokens
            start: Starting position

        Returns:
            (Visitable node, next position)
        """
        i = start

        # Skip opening bracket if present
        if i < len(tokens) and tokens[i].type == "LBRACKET":
            i += 1

        # Skip question mark if present
        if i < len(tokens) and tokens[i].type == "QUESTION":
            i += 1

        # Check for NOT operator (RFC 9535: !)
        has_not = False
        if i < len(tokens) and tokens[i].type == "NOT":
            has_not = True
            i += 1

        # Skip opening parenthesis if present
        if i < len(tokens) and tokens[i].type == "LPAREN":
            i += 1
            # Recursively parse expression inside parentheses
            node, i = self._parse_expression(tokens, i)
            # Skip closing parenthesis
            if i < len(tokens) and tokens[i].type == "RPAREN":
                i += 1
        else:
            # Parse left side (field access or nested wildcard)
            left_node, i = self._parse_field_access(tokens, i)

            # Check if left_node is a Wildcard (nested wildcard case)
            if isinstance(left_node, Wildcard):
                # This is a nested wildcard - return it directly
                node = left_node
            else:
                # Parse operator
                if i >= len(tokens):
                    return left_node, i

                op_token = tokens[i]
                i += 1

                # Parse right side (value)
                right_node, i = self._parse_value(tokens, i)

                # Create comparison node
                if op_token.type == "EQ":
                    node = Equal(left_node, right_node)
                elif op_token.type == "NE":
                    node = NotEqual(left_node, right_node)
                elif op_token.type == "GT":
                    node = GreaterThan(left_node, right_node)
                elif op_token.type == "LT":
                    node = LessThan(left_node, right_node)
                elif op_token.type == "GTE":
                    node = GreaterThanEqual(left_node, right_node)
                elif op_token.type == "LTE":
                    node = LessThanEqual(left_node, right_node)
                else:
                    raise SyntaxError(f"Unexpected operator: {op_token.type}")

            # Skip closing parenthesis if present (from earlier opening)
            if i < len(tokens) and tokens[i].type == "RPAREN":
                i += 1

        # Apply NOT if present
        if has_not:
            node = Not(node)

        # Check for AND/OR (RFC 9535: && and ||)
        if i < len(tokens) and tokens[i].type in ("AND", "OR"):
            op = tokens[i].type
            i += 1
            right_expr, i = self._parse_expression(tokens, i)
            if op == "AND":
                node = And(node, right_expr)
            else:
                node = Or(node, right_expr)

        return node, i

    def _parse_field_access(
        self, tokens: list[Token], start: int
    ) -> tuple[Visitable, int]:
        """
        Parse field access expression (including nested paths and wildcards).

        Supports:
        - Simple: @.field
        - Nested: @.a.b.c
        - Nested wildcard: @.items[*][?@.price > 100]

        Args:
            tokens: List of tokens
            start: Starting position

        Returns:
            (Field node or Wildcard node, next position)
        """
        i = start

        # Check for @ (current item)
        if i < len(tokens) and tokens[i].type == "AT":
            i += 1
            # Use Item() only in wildcard context, otherwise GlobalScope()
            parent = Item() if self._is_wildcard_context else GlobalScope()
        else:
            parent = GlobalScope()

        # Skip dot
        if i < len(tokens) and tokens[i].type == "DOT":
            i += 1

        # Parse field path chain (e.g., a.b.c)
        field_chain = []
        while i < len(tokens) and tokens[i].type == "IDENTIFIER":
            field_chain.append(tokens[i].value)
            i += 1

            # Check for dot (continues path)
            if i < len(tokens) and tokens[i].type == "DOT":
                # Check if next token is also an identifier
                if i + 1 < len(tokens) and tokens[i + 1].type == "IDENTIFIER":
                    i += 1  # Skip dot
                    continue
                else:
                    # Dot but no identifier after - break
                    break
            else:
                # No dot, break
                break

        if not field_chain:
            raise SyntaxError(f"Expected field name at position {i}")

        # Check for nested wildcard on last field: field[*][?...]
        if len(field_chain) > 0 and self._check_nested_wildcard(tokens, i):
            # Build parent chain for all fields except the last
            for field in field_chain[:-1]:
                parent = Object(parent, field)

            # Last field is the collection for the wildcard
            collection_name = field_chain[-1]
            return self._parse_nested_wildcard(tokens, i, parent, collection_name)

        # Build nested Field structure
        # e.g., ["a", "b", "c"] -> Field(Object(Object(parent, "a"), "b"), "c")
        for field in field_chain[:-1]:
            parent = Object(parent, field)

        # Last field
        field_name = field_chain[-1]
        return Field(parent, field_name), i

    def _check_nested_wildcard(self, tokens: list[Token], start: int) -> bool:
        """
        Check if tokens starting at position indicate a nested wildcard pattern.

        Pattern: [*][?...]

        Args:
            tokens: List of tokens
            start: Starting position

        Returns:
            True if nested wildcard pattern detected
        """
        i = start

        # Check for [*]
        if (
            i + 2 < len(tokens)
            and tokens[i].type == "LBRACKET"
            and tokens[i + 1].type == "WILDCARD"
            and tokens[i + 2].type == "RBRACKET"
        ):
            # Check if followed by [?...]
            if (
                i + 3 < len(tokens)
                and tokens[i + 3].type == "LBRACKET"
            ):
                return True

        return False

    def _parse_nested_wildcard(
        self, tokens: list[Token], start: int, parent: Visitable, collection_name: str
    ) -> tuple[Wildcard, int]:
        """
        Parse nested wildcard pattern: collection[*][?predicate]

        Args:
            tokens: List of tokens
            start: Position after collection name
            parent: Parent node (Item or GlobalScope)
            collection_name: Name of the collection field

        Returns:
            (Wildcard node, next position)
        """
        i = start

        # Skip [*]
        if (
            i + 2 < len(tokens)
            and tokens[i].type == "LBRACKET"
            and tokens[i + 1].type == "WILDCARD"
            and tokens[i + 2].type == "RBRACKET"
        ):
            i += 3
        else:
            raise SyntaxError(f"Expected [*] at position {i}")

        # Parse filter expression [?...]
        if i < len(tokens) and tokens[i].type == "LBRACKET":
            # Save current wildcard context
            old_context = self._is_wildcard_context

            # Set wildcard context to True for nested predicate
            self._is_wildcard_context = True
            predicate, i = self._parse_expression(tokens, i)

            # Restore previous context
            self._is_wildcard_context = old_context

            # Create Wildcard node
            collection_obj = Object(parent, collection_name)
            return Wildcard(collection_obj, predicate), i

        raise SyntaxError(f"Expected filter expression at position {i}")

    def _parse_value(self, tokens: list[Token], start: int) -> tuple[Value, int]:
        """
        Parse a value (literal or placeholder).

        Args:
            tokens: List of tokens
            start: Starting position

        Returns:
            (Value node, next position)
        """
        i = start

        if i >= len(tokens):
            raise SyntaxError("Expected value but reached end of tokens")

        token = tokens[i]

        if token.type == "NUMBER":
            # Parse number
            value = float(token.value) if "." in token.value else int(token.value)
            return Value(value), i + 1

        elif token.type == "STRING":
            # Parse string (remove quotes)
            value = token.value[1:-1]
            return Value(value), i + 1

        elif token.type == "PLACEHOLDER":
            # This is a placeholder - will be bound later
            # Return a special marker value
            value_node = self._create_placeholder_value(token.value)
            return value_node, i + 1

        elif token.type == "IDENTIFIER":
            # Could be a boolean literal
            if token.value.lower() == "true":
                return Value(True), i + 1
            elif token.value.lower() == "false":
                return Value(False), i + 1
            elif token.value.lower() == "null":
                return Value(None), i + 1

        raise SyntaxError(f"Unexpected token in value position: {token}")

    def _create_placeholder_value(self, placeholder_str: str) -> Value:
        """
        Create a placeholder value that will be bound later.

        Args:
            placeholder_str: Placeholder string (e.g., %d, %(name)s)

        Returns:
            Value node with placeholder marker
        """
        # We'll store a special marker that we'll replace during match()
        value = Value(("__PLACEHOLDER__", self._placeholder_bind_index))
        self._placeholder_bind_index += 1
        return value

    def _parse_path(self, tokens: list[Token]) -> tuple[Visitable, bool]:
        """
        Parse the full JSONPath expression (supports nested paths).

        Supports:
        - Simple: $.items[?@.price > 100]
        - Nested: $.store.items[?@.price > 100]
        - Deep nested: $.a.b.c.items[?@.x > 1]

        Args:
            tokens: List of tokens

        Returns:
            (Visitable node, is_wildcard)
        """
        i = 0

        # Skip $
        if i < len(tokens) and tokens[i].type == "DOLLAR":
            i += 1

        # Skip .
        if i < len(tokens) and tokens[i].type == "DOT":
            i += 1

        # Parse path chain (e.g., a.b.c)
        path_chain = []
        while i < len(tokens) and tokens[i].type == "IDENTIFIER":
            path_chain.append(tokens[i].value)
            i += 1

            # Check for dot (continues path)
            if i < len(tokens) and tokens[i].type == "DOT":
                i += 1
                # Continue to next identifier
            else:
                # No more dots, break
                break

        if not path_chain:
            # No path found, check if it's just a filter without path
            # e.g., $[?@.age > 25]
            if i < len(tokens) and tokens[i].type == "LBRACKET":
                # Simple filter without path
                self._is_wildcard_context = False
                predicate, _ = self._parse_expression(tokens, i)
                return predicate, False
            raise SyntaxError("Expected path or filter expression")

        # Build nested Object structure from path chain
        # e.g., ["a", "b", "c"] -> Object(Object(Object(GlobalScope(), "a"), "b"), "c")
        parent = GlobalScope()
        for path_element in path_chain[:-1]:
            parent = Object(parent, path_element)

        # Last element in path is the collection name
        collection_name = path_chain[-1]

        # Check for wildcard [*]
        is_wildcard = False
        if (
            i + 2 < len(tokens)
            and tokens[i].type == "LBRACKET"
            and tokens[i + 1].type == "WILDCARD"
            and tokens[i + 2].type == "RBRACKET"
        ):
            is_wildcard = True
            i += 3

        # Parse filter expression
        if i < len(tokens) and tokens[i].type == "LBRACKET":
            if is_wildcard:
                # Wildcard with filter
                self._is_wildcard_context = True
                predicate, _ = self._parse_expression(tokens, i)
                self._is_wildcard_context = False

                # Create Wildcard node
                collection_obj = Object(parent, collection_name)
                return Wildcard(collection_obj, predicate), True
            else:
                # Simple filter without wildcard
                self._is_wildcard_context = False
                predicate, _ = self._parse_expression(tokens, i)
                return predicate, False

        raise SyntaxError("Expected filter expression")

    def _bind_placeholder(
        self, value: Any, params: Union[Tuple[Any, ...], Dict[str, Any]]
    ) -> Any:
        """
        Bind a placeholder to its actual value.

        Args:
            value: Value (may contain placeholder marker)
            params: Parameter values

        Returns:
            Actual value
        """
        if isinstance(value, tuple) and len(value) == 2:
            marker, idx = value
            if marker == "__PLACEHOLDER__":
                if idx < len(self._placeholder_info):
                    ph_info = self._placeholder_info[idx]

                    # Get actual value from params
                    if ph_info["positional"]:
                        param_idx = int(ph_info["name"])
                        if param_idx < len(params):
                            return params[param_idx]
                    else:
                        if ph_info["name"] in params:
                            return params[ph_info["name"]]

                    # If not found, return marker as-is
                    return value

        return value

    def _bind_values_in_ast(
        self, node: Visitable, params: Union[Tuple[Any, ...], Dict[str, Any]]
    ) -> Visitable:
        """
        Recursively bind placeholder values in the AST.

        Args:
            node: AST node
            params: Parameter values

        Returns:
            AST node with bound values
        """
        if isinstance(node, Value):
            # Bind the value if it's a placeholder
            bound_value = self._bind_placeholder(node.value(), params)
            return Value(bound_value)

        elif isinstance(node, (Equal, NotEqual, GreaterThan, LessThan, GreaterThanEqual, LessThanEqual)):
            # Recursively bind left and right
            left = self._bind_values_in_ast(node.left(), params)
            right = self._bind_values_in_ast(node.right(), params)
            return type(node)(left, right)

        elif isinstance(node, (And, Or)):
            left = self._bind_values_in_ast(node.left(), params)
            right = self._bind_values_in_ast(node.right(), params)
            return type(node)(left, right)

        elif isinstance(node, Not):
            operand = self._bind_values_in_ast(node.operand(), params)
            return Not(operand)

        elif isinstance(node, Wildcard):
            # Recursively bind predicate
            predicate = self._bind_values_in_ast(node.predicate(), params)
            return Wildcard(node.parent(), predicate)

        # For other nodes (Field, Item, GlobalScope, Object), return as-is
        return node

    def match(
        self, data: Any, params: Union[Tuple[Any, ...], Dict[str, Any]] = ()
    ) -> bool:
        """
        Check if data matches the specification with given parameters.

        Args:
            data: The data object to check (must implement Context protocol)
            params: Parameter values (tuple for positional, dict for named)

        Returns:
            True if data matches the specification, False otherwise

        Examples:
            >>> spec = parse("$[?(@.age > %d)]")
            >>> user = DictContext({"age": 30})
            >>> spec.match(user, (25,))
            True
        """
        # Check if data implements Context protocol (has 'get' method)
        if not hasattr(data, "get") or not callable(getattr(data, "get")):
            raise TypeError(
                f"Data must implement Context protocol (have a 'get' method), "
                f"got {type(data).__name__}"
            )

        # Reset placeholder binding index
        self._placeholder_bind_index = 0

        # Tokenize
        lexer = Lexer(self.template)
        tokens = lexer.tokenize()

        # Parse to AST
        spec_ast, is_wildcard = self._parse_path(tokens)

        # Bind placeholder values
        bound_ast = self._bind_values_in_ast(spec_ast, params)

        # Evaluate using EvaluateVisitor
        visitor = EvaluateVisitor(data)
        bound_ast.accept(visitor)

        return visitor.result()


def parse(template: str) -> NativeParametrizedSpecification:
    """
    Parse RFC 9535 compliant JSONPath expression with C-style placeholders (native implementation).

    Args:
        template: JSONPath with %s, %d, %f or %(name)s placeholders

    Returns:
        NativeParametrizedSpecification that can be executed with different parameter values

    Examples:
        >>> spec = parse("$[?@.age > %d]")
        >>> user = DictContext({"age": 30})
        >>> spec.match(user, (25,))
        True

        >>> spec = parse("$[?@.name == %(name)s]")
        >>> user = DictContext({"name": "Alice"})
        >>> spec.match(user, {"name": "Alice"})
        True

        >>> spec = parse("$[?@.age > %d && @.active == %s]")
        >>> user = DictContext({"age": 30, "active": True})
        >>> spec.match(user, (25, True))
        True

        >>> spec = parse("$[?!(@.active == %s)]")
        >>> user = DictContext({"active": False})
        >>> spec.match(user, (True,))
        True
    """
    return NativeParametrizedSpecification(template)



import functools
import os
import time
import shlex
import shutil
import logging
import datetime
import subprocess

from psycopg.conninfo import conninfo_to_dict
from ascetic_ddd.faker.infrastructure.dump.interfaces import IDump, IFileDump

__all__ = ("SinglePgDump",)


class SinglePgDump(IFileDump):
    """
    https://github.com/IBM/backwork-backup-postgresql/blob/master/postgresql/postgresql.py
    """
    _credentials: dict
    _base_path: str
    _ttl: datetime.timedelta

    def __init__(self, base_path: str, ttl: datetime.timedelta, conninfo: str):
        self._base_path = base_path
        self._ttl = ttl
        self._logger = logging.getLogger(__name__)
        self._credentials = conninfo_to_dict(conninfo)

    @functools.cached_property
    def _cpu_count(self) -> int:
        cpu_count = len(os.sched_getaffinity(0))  # os.process_cpu_count() from 3.13v
        cpu_count = min(max(2, cpu_count), 99)  # max_connections
        return cpu_count

    @property
    def ttl(self) -> datetime.timedelta:
        return self._ttl

    async def exists(self, name: str) -> bool:
        filename = self.make_filepath(name)
        if os.path.exists(filename):
            mtime = os.path.getmtime(filename)
            delta = time.time() - mtime
            if delta < self._ttl.total_seconds():
                self._logger.info("Dump %s exists", filename)
                return True
        self._logger.info("Dump %s does not exist", filename)
        return False

    async def dump(self, name: str):
        if not os.path.exists(self._base_path):
            os.mkdir(self._base_path)
        filename = self.make_filepath(name)
        if os.path.exists(filename):
            shutil.rmtree(filename)
        os.environ["PGPASSWORD"] = self._credentials['password']
        cmd = " ".join([
            shlex.join([
                "pg_dump",
                "-c",
                "-Fd",
                "-j", str(self._cpu_count),
                "-h", self._credentials['host'],
                "-p", self._credentials['port'],
                "-U", self._credentials['user'],
                "-d", self._credentials['dbname'],
                "--no-password",
                # "--inserts",
                "--no-owner",
                "--no-privileges",
                "-Z", "9",
                "-f", shlex.quote(filename),
            ])
        ])
        proc = subprocess.Popen(cmd, shell=True, env={'PGPASSWORD': self._credentials['password']},
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        sub_out, sub_err = proc.communicate()
        code = proc.returncode
        if code != 0:
            raise Exception("pg_dump failed: %s, %s", proc.returncode, sub_err)

    async def load(self, name: str):
        os.environ["PGPASSWORD"] = self._credentials['password']
        """
        cmd = " ".join([
            shlex.join(["gunzip", "-c", shlex.quote(self._make_filename(name)),]),
            "|",
            shlex.join([
                "psql",
                "-h", self._credentials['host'],
                "-p", self._credentials['port'],
                "-U", self._credentials['user'],
                "-d", self._credentials['dbname'],
                "--no-password",
            ])
        ])
        """
        cmd = shlex.join([
            "pg_restore",
            "-c",
            "-j", str(self._cpu_count),
            "-h", self._credentials['host'],
            "-p", self._credentials['port'],
            "-U", self._credentials['user'],
            "-d", self._credentials['dbname'],
            "--no-password",
            "--no-owner",
            "--no-privileges",
            shlex.quote(self.make_filepath(name))
        ])
        proc = subprocess.Popen(cmd, shell=True, env={'PGPASSWORD': self._credentials['password']},
                                stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        sub_out, sub_err = proc.communicate()
        code = proc.returncode
        if code != 0:
            raise Exception("pg_restore failed: %s, %s", proc.returncode, sub_err)

    def make_filepath(self, name: str) -> str:
        return os.path.join(self._base_path, "%s.sql.gz" % name)



import typing

from ascetic_ddd.observable.interfaces import IObservable


__all__ = ('IPgExternalSource',)


@typing.runtime_checkable
class IPgExternalSource(IObservable, typing.Protocol):
    @property
    def table(self) -> str:
        ...



"""
Example usage of JSONPath RFC 9535 to SQLAlchemy SQL compiler.

Demonstrates how to compile RFC 9535 compliant JSONPath expressions
into SQLAlchemy Select queries.
"""
try:
    from sqlalchemy import MetaData, Table, Column, Integer, String, Numeric, Boolean, ForeignKey
    HAS_SQLALCHEMY = True
except ImportError:
    HAS_SQLALCHEMY = False
    print("SQLAlchemy is not installed. Install with: pip install sqlalchemy")

from ascetic_ddd.jsonpath_rfc9535_ext.infrastructure.jsonpath_to_sqlalchemy_sql import (
    JSONPathToSQLAlchemyCompiler, SchemaMetadata, RelationshipMetadata, RelationType
)


def create_example_schema() -> 'SchemaMetadata':
    """Create example schema with SQLAlchemy Tables."""
    if not HAS_SQLALCHEMY:
        raise ImportError("SQLAlchemy is required for this example")

    metadata = MetaData()

    # Define tables
    users = Table(
        'users',
        metadata,
        Column('id', Integer, primary_key=True),
        Column('name', String(100), nullable=False),
        Column('age', Integer),
        Column('email', String(255)),
        Column('active', Boolean, nullable=False),
    )

    orders = Table(
        'orders',
        metadata,
        Column('id', Integer, primary_key=True),
        Column('user_id', Integer, ForeignKey('users.id'), nullable=False),
        Column('total', Numeric(10, 2), nullable=False),
        Column('status', String(50), nullable=False),
    )

    products = Table(
        'products',
        metadata,
        Column('id', Integer, primary_key=True),
        Column('name', String(100), nullable=False),
        Column('price', Numeric(10, 2), nullable=False),
        Column('category', String(50)),
    )

    # Define relationships
    relationships = {
        'users': {
            'orders': RelationshipMetadata(
                target_table='orders',
                foreign_key='user_id',
                target_primary_key='id',
                relationship_type=RelationType.ONE_TO_MANY,
            ),
        },
    }

    schema = SchemaMetadata(
        tables={
            'users': users,
            'orders': orders,
            'products': products,
        },
        relationships=relationships,
        root_table='users',
    )

    return schema


def main():
    """Run examples."""
    if not HAS_SQLALCHEMY:
        print("SQLAlchemy is not installed. Install with: pip install sqlalchemy")
        return

    schema = create_example_schema()
    compiler = JSONPathToSQLAlchemyCompiler(schema)

    print("=" * 80)
    print("JSONPath RFC 9535 to SQLAlchemy SQL Compiler Examples")
    print("=" * 80)

    examples = [
        ("Simple field access", "$.name"),
        ("Wildcard (all columns)", "$[*]"),
        ("Filter on age", "$[?@.age > 18]"),
        ("Filter with equality", "$[?@.name == 'John']"),
        ("Navigate to orders", "$.orders[*]"),
        ("Navigate and filter", "$.orders[?@.total > 100]"),
        ("Logical AND", "$[?@.age > 25 && @.active == true]"),
        ("Logical OR", "$[?@.age < 18 || @.age > 65]"),
        ("Logical NOT", "$[?!(@.active == false)]"),
    ]

    for idx, (description, jsonpath) in enumerate(examples, 1):
        print(f"\n{idx}. {description}")
        print(f"   JSONPath: {jsonpath}")
        print(f"   SQLAlchemy Query:")
        try:
            query = compiler.compile(jsonpath)
            # Format query for display
            print(f"      {query}")
        except Exception as e:
            print(f"      Error: {e}")

    print("\n" + "=" * 80)
    print("RFC 9535 Compliance Examples")
    print("=" * 80)
    print("\nDemonstrating standard compliance:")

    rfc_examples = [
        ("Standard equality operator (==)", "$[?@.age == 30]"),
        ("Standard logical AND (&&)", "$[?@.age >= 18 && @.age <= 65]"),
        ("Standard logical OR (||)", "$[?@.age < 18 || @.age > 65]"),
        ("Standard NOT operator (!)", "$[?!(@.active == true)]"),
    ]

    for idx, (description, jsonpath) in enumerate(rfc_examples, 1):
        print(f"\n{idx}. {description}")
        print(f"   JSONPath: {jsonpath}")
        print(f"   SQLAlchemy Query:")
        try:
            query = compiler.compile(jsonpath)
            print(f"      {query}")
        except Exception as e:
            print(f"      Error: {e}")

    print("\n" + "=" * 80)
    print("\nNote: SQLAlchemy queries can be executed like:")
    print("  with engine.connect() as conn:")
    print("      result = conn.execute(query)")
    print("=" * 80)


if __name__ == "__main__":
    main()
